/*
  BDE.c : BDE score function
  Copyright (C) 2009, 2013 Yoshinori Tamada <tamada  A T  is.s.u-tokyo.ac.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

/*!\defgroup BDE
 *\brief BDE score function
 *
 *\author Yoshinori Tamada &lt;tamada a t is.s.u-tokyo.ac.jp&gt;
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "BDE.h"
#include "io/ytFile.h"
#define BDE_PRIOR_TYPE_FIXED 0




/* Prototype declaration of the internal functions */
BDEBuff * BDEBuff_new();
int BDE_count_n_ijk(BDEBuff * bf, const int * parents, int q, int i);
//paraos_real_t BDE_calc_bde(int * n_ijk,	int i, const int * parents, int q, int count, int * r);
//paraos_real_t BDE_calc_bde(BDEElm ** n_ijk,	int i, const int * parents, int q, int count, int r);
float BDE_calc_bde(BDEElm ** n_ijk, const int count, const int r);
/*****************************************************************************/
/* M A I N   R O U T I N E S                                                 */
/*****************************************************************************/

/*!\brief Initializes the BDE score function.
 *
 *\param X  (\a n, \a p)-fortran-style input data matrix for parents.
 *\param Y  (\a n, \a p)-fortran-style input data matrix for children.
 *\param p  The number of nodes.
 *\param n  The number of samples.
 *\param r  The number of categories (types of discretized values) in data.
 *\param m  The number of maximum parents.
 *
 *\return  Pointer to the score working buffer.
 *\ingroup BDE */
inline void * BDE_init(int * X, int * Y, int p, int n, int r, int m, FILE * fp){
  
  BDEBuff * bf = BDEBuff_new();
  size_t total = 0;
  size_t mem = sizeof(BDEBuff); 
  total += mem;

  bf->p = p;
  bf->n = n;
  bf->X = X;
  bf->Y = Y;

  bf->max_parents = m;
  bf->verbose = 0;
  bf->err = fp;
  /*
  if(bf->verbose >= 0){
	fprintf(bf->err, "BDE: Initializing the BDe score function.\n");
  }
  */
  /* ------------------------------------------------------------- */
  /* Categories : bf->r_i                                          */
  /* ------------------------------------------------------------- */

  //mem = sizeof(int) * bf->p; 
  //total += mem;
  //bf->r = (int *) malloc(mem);

  //int max_r_i = 0;

  /* The program supports different numbers of categories for
	 variables.  Here the same number of categories is assumed.
	 You can change this by yourself. */
  //for(int i = 0; i < bf->p; i++){
  //bf->r[i] = c;
  //if(bf->r[i] > max_r_i) max_r_i = bf->r[i];
  //}/* for i < bf->n */

  /* This implementation does not support different numbers of
     categories for variables. */
  bf->r = r;

  /* ------------------------------------------------------------- */
  /* n_ij & n_ijk                                                  */
  /* ------------------------------------------------------------- */
  /* 
	 Note: The max num of patterns to count is equal to the number of
	 samples because if the patterns in all the samples are different,
	 then all the parent status patterns are different and they are
	 the entire possible patterns from the data.
  */
  /*
  mem = sizeof(int) * bf->max_parents * bf->n;
  total += mem;
  bf->n_ij = malloc(mem);
  mem = sizeof(int) * bf->n * max_r_i;
  total += mem;
  bf->n_ijk = malloc(mem);
  total += mem;
  mem = sizeof(int) * m;
  total += mem;
  bf->j = malloc(mem);
  */

  size_t rr = 1;
  size_t sum = 0;
  for(int i = 0; i <= m; i++){
	if(rr < n){
	  sum += rr;
	  rr *= r;
	} else {
	  sum += n;
	}
  }
  mem = sizeof(BDEElm) * sum * r;
  total += mem;
  bf->table = malloc(mem);
  //bf->max_tpos = 0;

  mem = sizeof(BDEElm * ) * bf->n;
  total += mem;
  bf->n_ijk = malloc(mem);

  /* Output summary */
  /*
  if(bf->verbose >= 0){
	fprintf(bf->err, "BDE: Total memory usage: %zu [bytes]", total);
	fprintf(bf->err, " (%zu [MiB])\n", (total + 524288) / 1048576);
	fprintf(bf->err, "BDE: Finished BDE initialization.\n");
	fflush(bf->err);
  }
  */
  fprintf(fp, "BDE: ");
  ytFile_printComma_size_t(fp, total);
  fprintf(fp, " bytes used.\n");
  return bf;
}
/* BDE_init() ============================================================== */

inline float BDE_score(const void * buff, const int j, const int * parents, const int q){

  BDEBuff * bf = (BDEBuff *) buff;

  /* likelihood */
  //Timer_start(bf->timer_1);
  int count = BDE_count_n_ijk(bf, parents, q, j);
  //Timer_stop(bf->timer_1);
  //#ifdef SIGN_DEBUG
  //fprintf(bf->err, "BDE_score(): count = %d\n", count);
  //#endif
  //Timer_start(bf->timer_2);
  //paraos_real_t score = BDE_calc_bde(bf->n_ijk, j, parents, q, count, bf->r);
  //fprintf(bf->err, "BDE_score(): score=%f\n", score);
  //return -2 * BDE_calc_bde(bf->n_ijk, j, parents, q, count, bf->r);
  return -2 * BDE_calc_bde(bf->n_ijk, count, bf->r);
  //Timer_stop(bf->timer_2);
  /* prior */
  /*
  switch(bf->prior){
  case BDE_PRIOR_TYPE_FIXED:
	score = -2 * score;
	break;

  default:
	fprintf(bf->err, "FATAL ERROR: BDE: Unknown prior type specified.\n");
	exit(1);
  }
  */
  //return score;
}
/* BDE_score() ============================================================= */

/*!\brief Releases memory allocated for the working buffer.
 *
 *\param buff Pointer to the working buffer returned by BDE_init().
 */
void BDE_finalize(void * buff){
  BDEBuff * bf = (BDEBuff *) buff;

  //free(bf->n_ij);
  free(bf->n_ijk);
  //free(bf->r);
  //free(bf->j);
  free(bf->table);

  free(buff);
}
/* BDE_finalize() ========================================================== */

/*****************************************************************************/
/* S U B   R O U T I N E S                                                   */
/*****************************************************************************/

inline BDEBuff * BDEBuff_new(){
  BDEBuff * bf = (BDEBuff *) malloc(sizeof(BDEBuff));

  bf->prior = BDE_PRIOR_TYPE_FIXED;
  bf->fixed_alpha = 1.0;
  //PROF
  //bf->timer_1 = Timer_new();
  //bf->timer_2 = Timer_new();
  return bf;
}
/* BDEBuff_new() =========================================================== */

int BDE_count_n_ijk(BDEBuff * bf, const int * parents, int q, int i){

  int n_ij_count = 0;

  bf->tpos = bf->r;

  if(q == 0){
	BDEElm * elm = bf->table;
    for(int k = 0; k < bf->r; k++) elm[k].value = 0.0;
    for(int l = 0; l < bf->n; l++){
      elm[bf->Y[l + i * bf->n]].value++;
    }
    bf->n_ijk[0] = elm;
    return 1;
  }

  for(int k = 0; k < bf->r; k++) bf->table[k].elm = NULL;
  for(int l = 0; l < bf->n; l++){
	/* search pattern */
	BDEElm * elm = bf->table;
	int j = 0;
    while(j < q){
	  BDEElm * next = elm[bf->X[l + parents[j] * bf->n]].elm;
	  if(next == NULL) break;
	  elm = next;
      j++;
	}

	if(j != q){
	  /* not found */
	  for(; j < q; j++){
		BDEElm * next = bf->table + bf->tpos;
		elm[bf->X[l + parents[j] * bf->n]].elm = next;
		for(int k = 0; k < bf->r; k++) next[k].elm = NULL;
		elm = next;
		bf->tpos += bf->r;
	  }
	  bf->n_ijk[n_ij_count++] = elm;
	}

	int xi = bf->Y[l + i * bf->n];
	elm[xi].value++;
  }/* for l < n */

  return n_ij_count;
}/* BDE_count_n_ijk() ===================================================== */
/*
inline paraos_real_t BDE_calc_bde(BDEElm ** n_ijk, int i, const int * parents, int q, int count, int r){

  paraos_real_t score = count * lgammaf(r);
  for(int j = 0; j < count; j++){
    int nij = 0;
    for(int k = 0; k < r; k++){
      int v = n_ijk[j][k].value;
      nij += v;
      score += lgammaf(1.0 + v);
    }
    score -= lgammaf(nij + r);
  }
  return score;
}
*/

inline float BDE_calc_bde(BDEElm ** n_ijk, const int count, const int r){

  //fprintf(stderr, "BDE: count=%d\n", count);

  float score = count * lgammaf(r);
  for(int j = 0; j < count; j++){
    int nij = 0;
    for(int k = 0; k < r; k++){
      int v = n_ijk[j][k].value;
      nij += v;
      score += lgammaf(1.0 + v);
      //fprintf(stderr, "BDE: k=%d, v=%d, nij=%d\n", k, v, nij);
      //fflush(stderr);
    }
    score -= lgammaf(nij + r);
  }
  return score;
}

/* BDE_calc_bde() ========================================================== */

/* End of file.  */
