/*
  ytCombo.{h,c} : Combination functions.
  Copyright (C) 2016, Yoshinori Tamada <tamada  A T  ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/
#include "ytCombo.h"
#include <stdlib.h>

/*!\brief Calculates the number of combinations of \a n things taken \a k at
 *  a time (binomial coefficient).
 *
 * This is the C implementation of ACM Algorithm 160.
 *
 *\param n
 *\param k
 *\return  number of combinations of \a n things taken \a k at a time
 *            (binomical coefficient for \a n and \a k).
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
uint64_t ytCombo_choose(int n, int k){
  int p = n - k, i;
  uint64_t r;
  if(k < p){
	p = k;
	k = n - p;
  }
  if(p == 0){
	return 1;
  }
  r = k + 1;
  for(i = 2; i <= p; i++){
	r = (r * (k + i)) / i;
  }
  return r;
}
/* ytCombo_choose() ======================================================== */

/*\!brief Calculates the (\a n,\a k) - binomical coefficient using
 * the index table.
 *
 *\param n
 *\param k
 *\param table index table calculated by ytCombo_indexTable().
 *\param tn
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
uint64_t ytCombo_chooseTable(int n, int k, const uint64_t * table, int tn){
  if (n == k || k == 0) {
    return 1;
  }
  if (n < k) {
    return 0;
  }
  if (n == k + 1) {
    return n;
  }
  return table[(n - k - 1) * tn + k + 1];
}/* ytCombo_chooseTable() ================================================== */

/*!\brief Generates the combination.
 *
 * This generates the combination of \a n things taken \a k at a time
 * in lexicographical order from the previouly generated combination.
 *
 * To generate the first combination, set -1 for the first positions of \a ar.
 *
 *\param n
 *\param k
 *\param ar
 *\return 1 if the generated combination is the last one, 0 otherwise.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
int ytCombo_generate(int n, int k, int * ar){
  int i, j;

  if(ar[0] == -1){
	for(i = 0; i < k; i++) ar[i] = i;
	return 0;
  }

  /* Increment the elements from the last position. */
  for(i = 0; i < k; i++){
	ar[k - i - 1]++;

	/* If the incremented element did not become the maximum value... */
	if(ar[k - i - 1] != n - i){
	  /* update the lower digits */
	  for(j = 0; j < i; j++){
		ar[k - i + j] = ar[k - i - 1] + j + 1;
	  }
	  
	  return 0; /* OK */
	}/* if () */
  }/* for i */
  return 1; /* Finished */
}/* ytCombo_generate() ===================================================== */

/*!\brief Generates the index table for ytCombo routines.
 *
 * The table has to have size <TT>sizeof(size_t) * n * n</TT>.
 *
 *\param n
 *\param table pointer to room for storing the index table.
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
void ytCombo_indexTable(int n, uint64_t * table){
  int i, j;

  for(j = 0; j < n; j++) table[j] = j;
  
  for(i = 1; i < n; i++){
	size_t prev = 0;
	table[i * n] = prev;
	
	for(j = 1; j < n; j++){
	  prev = prev + table[(i - 1) * n + j];
	  table[i * n + j] = prev;
	}/* for j < n */
  }/* for i < n */
}
/* ytCombo_indexTable() ====================================================== */

/*!\brief Calculates the index for the given combination.
 *
 * Note that this returns the reverse order (index) of the combination.
 * To obtain the normal index, calculate (choose(n, c) - \t id) where
 * \t id is the reverse index returned by this function.
 *
 *\param ar sorted list (combination) of values.
 *\param table index table calculated by ytCombo_index_table().
 *\param tn table size used to calculate \a table by ytCombo_index_table().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
uint64_t ytCombo_index(int n, int c, const int * ar, const uint64_t * table, 
                       int tn){
  int i;
  uint64_t ret = 0;
  int n_c = n - c;
  
  for(i = 0; i < c; i++){
	int v = n_c - ar[i] + i; /* = n - ar[i] - 1 - (c - i - 1) */
	ret += table[(c - i - 1) * tn + v];
  }
  return ret;
}
/* ytCombo_index() ========================================================= */

uint64_t ytCombo_index2(int n, int c, const int * ar, const int * a,
                      const uint64_t * table, int tn){
  int i;
  uint64_t ret = 0;
  int n_c = n - c;
  
  for(i = 0; i < c; i++){
	int v = n_c - a[ar[i]] + i; /* = n - ar[i] - 1 - (c - i - 1) */
	ret += table[(c - i - 1) * tn + v];
  }
  return ret;
}
/* ytCombo_index2() ======================================================== */

uint64_t ytCombo_index3(int n, int c, const int * ar, const int * a, 
                        const int * b, const uint64_t * table, int tn){
  int i;
  uint64_t ret = 0;
  int n_c = n - c;
  
  for(i = 0; i < c; i++){
	int v = n_c - b[a[ar[i]]] + i; /* = n - ar[i] - 1 - (c - i - 1) */
	ret += table[(c - i - 1) * tn + v];
  }
  return ret;
}
/* ytCombo_index3() ======================================================== */

/*!\brief Obtain the combination corresponding to the given index.
 *
 * This is the inverse function of ytCombo_index().
 *
 *\param tn table size specified to ytCombo_indexTable().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
void ytCombo_combination(int n, int k, uint64_t index, int * ar, 
                         const uint64_t * table, int tn){
  int i;
  uint32_t right = n - k;
  uint32_t p = (k - 1) * tn;

  for (i = 0; i < k; i++) {
    uint32_t left = 0;
    right++;
    
    while (left < right) {
      uint32_t mid = (left + right) >> 1;

      if (table[p + mid] <= index) {
        left = mid + 1;
      }else{
        right = mid;
      }
    }
    left--;
    ar[i] = n - left - k + i;
    index -= table[p + left];
    right = left;
    p -= tn;
  }
}
/* ytCombo_combination() =================================================== */

/*!\brief Obtain the combination corresponding to the given index.
 *
 * This is the inverse function of ytCombo_index().
 *
 *\param tn table size specified to combo_indexTable().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
void ytCombo_combination2(int n, int k, uint64_t index, int * ar, 
                          const int * a, const uint64_t * table, int tn){
  int i;
  uint32_t right = n - k;
  uint32_t p = (k - 1) * tn;

  for (i = 0; i < k; i++) {
    uint32_t left = 0;
    right++;
    
    while (left < right) {
      uint32_t mid = (left + right) >> 1;

      if (table[p + mid] <= index) {
        left = mid + 1;
      }else{
        right = mid;
      }
    }
    left--;
    ar[i] = a[n - left - k + i];
    index -= table[p + left];
    right = left;
    p -= tn;
  }
}
/* ytCombo_combination2() ================================================== */

/*!\brief Obtain the combination corresponding to the given index.
 *
 * This is the inverse function of ytCombo_index().
 *
 *\param tn table size specified to combo_indexTable().
 *
 * <DIV align="right"><A href="#top">Go to top</A></DIV>
 *\ingroup ytCombo */
void ytCombo_combination3(int n, int k, uint64_t index, int * ar, 
                          const int * a, const int * b, 
                          const uint64_t * table, int tn){
  int i;
  uint32_t right = n - k;
  uint32_t p = (k - 1) * tn;

  for (i = 0; i < k; i++) {
    uint32_t left = 0;
    right++;
    
    while (left < right) {
      uint32_t mid = (left + right) >> 1;

      if (table[p + mid] <= index) {
        left = mid + 1;
      }else{
        right = mid;
      }
    }
    left--;
    ar[i] = b[a[n - left - k + i]];
    index -= table[p + left];
    right = left;
    p -= tn;
  }
}
/* ytCombo_combination3() ================================================== */

void ytCombo_print(FILE * fp, size_t n, const int * A){
  fprintf(fp, "{");
  if(n != 0) fprintf(fp, "%d", A[0]);
  for(size_t i = 1; i < n; i++){
    fprintf(fp, ",%d", A[i]);
  }
  fprintf(fp, "}");
}/* ytCombo_print() ======================================================== */

void ytCombo_print2(FILE * fp, size_t n, const int * A, const int * B){
  fprintf(fp, "{");
  if(n != 0) fprintf(fp, "%d", B[A[0]]);
  for(size_t i = 1; i < n; i++){
    fprintf(fp, ",%d", B[A[i]]);
  }
  fprintf(fp, "}");
}/* ytCombo_print2() ======================================================= */

void ytCombo_bitInit(uint64_t * bit, int k){
  *bit = 0;
  for(int i = 0; i < k; i++){
    *bit <<= 1;
    *bit |= 1;
  }
}/* ytCombo_bitInit() ====================================================== */

void ytCombo_bitNext(uint64_t * bit){
  uint64_t smallest = *bit & -*bit;
  uint64_t ripple = *bit + smallest;
  uint64_t new_smallest = ripple & -ripple;
  uint64_t ones = ((new_smallest / smallest) >> 1) - 1;
  *bit = ripple | ones;
}/* ytCombo_bitNext() ====================================================== */

void ytCombo_bit(int n, int k, uint64_t index, uint64_t * bit,
                 const uint64_t * table, int tn){
  int i;
  uint32_t right = n - k;
  uint32_t p = (k - 1) * tn;
  *bit = 0;

  for (i = 0; i < k; i++) {
    uint32_t left = 0;
    right++;
    
    while (left < right) {
      uint32_t mid = (left + right) >> 1;

      if (table[p + mid] <= index) {
        left = mid + 1;
      }else{
        right = mid;
      }
    }
    left--;
    //ar[i] = n - left - k + i;
    *bit |= 1 << (left + k - i - 1);
    index -= table[p + left];
    right = left;
    p -= tn;
  }
}/* ytCombo_bitt() ========================================================= */

void ytCombo_bitCombo2(int n, int k, uint64_t index,
                       uint64_t * bit, int * A, const int * a,
                       const uint64_t * table, int tn){
  int i;
  uint32_t right = n - k;
  uint32_t p = (k - 1) * tn;
  *bit = 0;

  for (i = 0; i < k; i++) {
    uint32_t left = 0;
    right++;
    
    while (left < right) {
      uint32_t mid = (left + right) >> 1;

      if (table[p + mid] <= index) {
        left = mid + 1;
      }else{
        right = mid;
      }
    }
    left--;
    int ii = n - left - k + i;
    A[i] = a[ii];
    *bit |= 1 << (n - ii - 1);
    index -= table[p + left];
    right = left;
    p -= tn;
  }
}/* ytCombo_bitCombo2()===================================================== */

void ytCombo_printBit(FILE * fp, uint64_t bit, int n){
  uint64_t filter = 1 << (n - 1);
  for(int i = 0; i < n; i++){
    if(bit & filter){
      fprintf(fp, "1");
    } else {
      fprintf(fp, "0");
    }
    filter >>= 1;
  }
}/* ytCombo_printBit() ===================================================== */

void ytCombo_printBitCombo(FILE * fp, uint64_t bit, int n){
  uint64_t filter = 1 << (n - 1);
  fprintf(fp, "{");
  int c = 0;
  for(int i = 0; i < n; i++){
    
    if(bit & filter){
      if(c != 0) fprintf(fp, ",");      
      fprintf(fp, "%d", i);
      c++;
    }
    filter >>= 1;
  }
  fprintf(fp, "}");
}/* ytCombo_printBitCombo() ================================================ */

void ytCombo_test(FILE * fp){

  int n = 10;
  int k = 4;
  uint64_t combo = 0;
  uint64_t * table = malloc(sizeof(n * n));
  ytCombo_indexTable(n, table);

  /* generate the initial bit combination. */
  for(int i = 0; i < k; i++){
    combo |= 1 << i;
  }

  uint64_t total = ytCombo_choose(n, k);

  int A[k];
  uint64_t bit;

  for(uint64_t i = 0; i < total; i++){
    //ytCombo_printBit(fp, combo, n);
    ytCombo_printBitCombo(fp, combo, n);

    ytCombo_combination(n, k, i, A, table, n);
    ytCombo_print(fp, k, A);

    ytCombo_bit(n, k, i, &bit, table, n);
    ytCombo_printBitCombo(fp, bit, n);

    fprintf(fp, "\n");

    ytCombo_bitNext(&combo);
  }
}/* ytCombo_test() ========================================================= */



/* End of file. */
