/*
  os.c : Original sequential optimal search algorithm.
  Copyright (C) 2009, 2013, 2015, 2016 
            Yoshinori Tamada <tamada  A T  is.s.u-tokyo.ac.jp>
  All rights reserved.

  This is the implementation of the OS algorithm by Ott, S. et al (2004).

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "paraos.h"
#include "os.h"
#include "Util.h"

#include "score.h"

#include "ytLib.h"
#include "util/ytTimer.h"
#include "io/ytFile.h"
#include "math/ytCombo.h"

#include <float.h>
#include <string.h>

#ifdef _OPENMP
#include <omp.h>
#else
#include "fakeomp.h"
#endif

typedef struct {
  double sc; /* score */
  size_t B; /* reverse lexicographic index in V - {g} of the best set of 
			   parents for 'g' */
  int len;  /* number of parents */
} FDB;

typedef struct {
  double sc; /* score */
  int * pi;  /* order */
  size_t * idx; /* optimal subsets for each node */
  int * len; /* optimal subset size */
} QDB;

void os_memcheck(void * p, char * msg);
void os_prepare(int p, FDB **** fdb, QDB *** qdb, 
				uint64_t ** table, int *** wog, int *** rwog);
size_t os_prepare_Bm(int **** Bm, int p);
void os_exec(void ** sb, int p, int max_parents,
			 FDB *** fdb, QDB ** qdb, 
			 uint64_t * table, int ** wog, int ** rwog);
FDB os_F(void * sb, int g, int * A, int a, FDB * fdb, int p,
		 int * wog, uint64_t * table, int max_parents);
void os_calc_F(void ** sb, int p, int max_parents, int a, 
			   FDB *** fdb, uint64_t * table, int ** wog, int ** rwog);
void os_calc_F_mp(void ** sb, const int p, int max_parents, int a, 
				  FDB *** fdb, uint64_t * table, int ** wog, int ** rwog);
void os_calc_Q(int a, FDB *** fdb, QDB ** qdb, int p, uint64_t * table, int ** wog);
void os_Q(int * A, int a, FDB ** fdb, QDB * qdb, int n, int ** wog, 
		  uint64_t * table, int tn, QDB * ret);
ytPCGraph * os_final_graph(int p, QDB * qdb, uint64_t * table);



ytPCGraph * os(Args * args, int p, void ** sb){
  /*
	fdb[a][g][i] where 
	    a \in {0, 1} is a DP step layer,
		g is a target gene, and
		i is a reverse lexicographic index of the combination pattern in V - {g}.

	We require two DP layers to use the results of the previous step
	and to store the results of the current step.
   */
  FDB *** fdb;

  /*
	qdb[a][i] where
	    a \in {0, 1} is a DP step layer, and
		i is a reverse lexicographic index of the combination patter in V.
   */
  QDB ** qdb;
  uint64_t * table;
  int ** wog;
  int ** rwog;
  ytPCGraph * g = NULL;
  ytTimer * timer = ytTimer_new();
  int max_parents = p - 1;

  /* Bm[x][y][z] represents the z-th element of the y-th combination vector
	 of x - 1 elements taken from x elements. */
  //int *** Bm;
  
  ytTimer_start(timer);

  if(args->MPI_rank == 0){
    os_prepare(p, &fdb, &qdb, &table, &wog, &rwog);
    //size_t total = os_prepare_Bm(&Bm, p);
    //ytFile_printComma_size_t(PARAOS_ERR, total);
    //fprintf(PARAOS_ERR, " bytes used for Bm.\n");
    //os_exec(sf, sb, p, max_parents, fdb, qdb, table, wog, rwog, Bm);
    os_exec(sb, p, max_parents, fdb, qdb, table, wog, rwog);
    g = os_final_graph(p, qdb[(p + 1) % 2], table);
  }

  ytTimer_stop(timer);

  fprintf(ytLib_ERR, "ALGO TIME: ");
  ytTimer_print_DDHHMMSS(timer, ytLib_ERR);
  fprintf(ytLib_ERR, "\n%f\tsec\n", ytTimer_get(timer));

  return g;
}
/* os() ==================================================================== */

void os_prepare(int p, FDB **** fdb, QDB *** qdb, 
                uint64_t ** table, int *** wog, int *** rwog){
  size_t mem;
  size_t total = 0;
  size_t max_pat;

  fprintf(ytLib_ERR, 
		  "Preparing the sequential version of the OS algorithm.\n");
  fflush(ytLib_ERR);
  
  /* ------------------------------------------------------------- */
  /* Index table                                                   */
  /* ------------------------------------------------------------- */
  mem = sizeof(size_t) * p * p;
  total += mem;
  fprintf(ytLib_ERR, "Index table size = ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes\n");
  (*table) = malloc(mem);
  os_memcheck(table, "table");
  ytCombo_indexTable(p, *table);

  /* FDB */
  fprintf(ytLib_ERR, "Allocating memory for F() ...\n");
  fprintf(ytLib_ERR, "  Allocating fdb.\n");
  fprintf(ytLib_ERR, "    sizeof(FDB**) = %zu\n", sizeof(FDB**));
  mem = sizeof(FDB**) * 2;
  fprintf(ytLib_ERR, "    total ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  (*fdb) = malloc(mem);
  os_memcheck(fdb, "fdb");
  total += mem;

  fprintf(ytLib_ERR, "  Allocating fdb[i].\n");
  fprintf(ytLib_ERR, "    sizeof(FDB*) = %zu\n", sizeof(FDB*));
  mem = sizeof(FDB*) * 2 * p;
  fprintf(ytLib_ERR, "    total ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  (*fdb)[0] = malloc(mem);
  os_memcheck(fdb[0], "fdb[i]");
  (*fdb)[1] = (*fdb)[0] + p;
  total += mem;

  fprintf(ytLib_ERR, "  Allocating fdb[i][j].\n");
  fprintf(ytLib_ERR, "    sizeof(FDB) = %zu\n", sizeof(FDB));
  max_pat = ytCombo_choose(p - 1, (p - 1) / 2);
  fprintf(ytLib_ERR, "  Max ");
  ytFile_printComma_size_t(ytLib_ERR, max_pat);
  fprintf(ytLib_ERR, " scores to save in one layer for each parent.\n");
  mem = max_pat * sizeof(FDB) * p * 2;
  fprintf(ytLib_ERR, "  ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  total += mem;
  (*fdb)[0][0] = malloc(mem);
  os_memcheck((*fdb)[0][0], "fdb");
  mem = 0;
  for(int i = 0; i < p; i++){
	(*fdb)[0][i] = (*fdb)[0][0] + mem;
	mem += max_pat;
	(*fdb)[1][i] = (*fdb)[0][0] + mem;
	mem += max_pat;
  }

  /* QDB */
  fprintf(ytLib_ERR, "Allocating memory for Q() ...\n");
  fprintf(ytLib_ERR, "  Allocating qdb.\n");
  fprintf(ytLib_ERR, "    sizeof(QDB*) = %zu.\n", sizeof(QDB*));
  mem = sizeof(QDB*) * 2;
  fprintf(ytLib_ERR, "    total ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  (*qdb) = malloc(mem);
  os_memcheck(*qdb, "qdb");
  total += mem;

  max_pat = ytCombo_choose(p, p / 2);
  fprintf(ytLib_ERR, "  Max ");
  ytFile_printComma_size_t(ytLib_ERR, max_pat);
  fprintf(ytLib_ERR, " patterns to store.\n");
  mem = max_pat * sizeof(QDB) * 2;
  fprintf(ytLib_ERR, "  ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  total += mem;
  (*qdb)[0] = malloc(mem);
  os_memcheck((*qdb)[0], "qdb");
  (*qdb)[1] = (*qdb)[0] + max_pat;

  fprintf(ytLib_ERR, "  Allocating qdb.pi for Q() ...\n");
  mem = sizeof(int) * p * max_pat * 2;
  fprintf(ytLib_ERR, "    ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  (*qdb)[0][0].pi = malloc(mem);
  os_memcheck((*qdb)[0][0].pi, "qdb.pi");
  total += mem;

  fprintf(ytLib_ERR, "  Allocating qdb.idx for Q() ...\n");
  mem = sizeof(size_t) * max_pat * p * 2;
  fprintf(ytLib_ERR, "    ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  (*qdb)[0][0].idx = malloc(mem);
  os_memcheck((*qdb)[0][0].idx, "qdb.idx");

  fprintf(ytLib_ERR, "  Allocating qdb.len in Q() ...\n");
  mem = sizeof(int) * max_pat * p * 2;
  fprintf(ytLib_ERR, "    ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " bytes.\n");
  (*qdb)[0][0].len = malloc(mem);
  os_memcheck((*qdb)[0][0].len, "qdb.len");
  total += mem;

  mem = 0;
  for(size_t i = 0; i < max_pat; i++){
	(*qdb)[0][i].pi = (*qdb)[0][0].pi + mem;
	(*qdb)[0][i].idx = (*qdb)[0][0].idx + mem;
	(*qdb)[0][i].len = (*qdb)[0][0].len + mem;
	mem += p;
	(*qdb)[1][i].pi = (*qdb)[0][0].pi + mem;
	(*qdb)[1][i].idx = (*qdb)[0][0].idx + mem;
	(*qdb)[1][i].len = (*qdb)[0][0].len + mem;
	mem += p;
  }

  /* wog: V => V - {j}, rwog: V - {j} => V */
  //int wog[p][p];
  //int rwog[p][p - 1];
  
  mem = sizeof(int *) * p * 2;
  (*wog) = malloc(mem);
  os_memcheck(*wog, "wog & rwog");
  (*rwog) = (*wog) + p;
  total += mem;

  mem = sizeof(int) * (p * p + p * (p - 1));
  int * ptr = malloc(mem);
  os_memcheck(ptr, "wog[i] & rwog[i]");
  total += mem;
  
  for(int j = 0; j < p; j++){
	int id = 0;
	(*wog)[j] = ptr;
	ptr += p;
	(*rwog)[j] = ptr;
	ptr += (p - 1);

	for(int k = 0; k < p; k++){
	  if(j == k){
		(*wog)[j][k] = -1;
	  } else {
		(*rwog)[j][id] = k;
		(*wog)[j][k] = id++;
	  }
	}
  }/* for j < p */
  
  fprintf(ytLib_ERR, "==========================================\n");
  fprintf(ytLib_ERR, " TOTAL MEMORY: ");
  ytFile_printComma_size_t(ytLib_ERR, total);
  fprintf(ytLib_ERR, " bytes.\n");
  fprintf(ytLib_ERR, "==========================================\n");
}
/* os_prepare() ============================================================ */

/*
  Generates all the combinations of (x - 1) elements taken from x elements
  where x is any value less than or equal to p.
  (*Bm)[x][y][z] represents the z-th element of the y-th combination
  of (x - 1) elements taken from x elements.

  This returns the total memory allocated.
 */
size_t os_prepare_Bm(int **** Bm, int p){
  size_t total = 0;
  size_t mem;
  int B[p];

  mem = sizeof(int **) * (p + 1); total += mem;
  *Bm = malloc(mem);
  for(int x = 2; x <= p; x++){
	mem = sizeof(int *) * x; total += mem;
	(*Bm)[x] = malloc(mem);
	mem = sizeof(int) * x * (x - 1); total += mem;
	int * ptr = malloc(mem);
	B[0] = -1;
	for(int y = 0; y < x; y++){
	  (*Bm)[x][y] = ptr; ptr += (x - 1);
	  ytCombo_generate(x, x - 1, B);
	  memcpy((*Bm)[x][y], B, sizeof(int) * (x - 1));
	}
  }

#ifdef PARAOS_DEBUG
  for(int x = 2; x <= p; x++){
	for(int y = 0; y < x; y++){
	  fprintf(ytLib_ERR, "{");
	  for(int z = 0; z < x - 1; z++){
		if(z != 0) fprintf(ytLib_ERR, ",");
		fprintf(ytLib_ERR, "%d", (*Bm)[x][y][z]);
	  }
	  fprintf(ytLib_ERR, "}\n");
	}
  }
#endif
  return total;
}
/* os_prepare_Bm() ========================================================= */

void os_exec(void ** sb, int p, int max_parents,
			 FDB *** fdb, QDB ** qdb, 
			 uint64_t * table, int ** wog, int ** rwog){
			 //			 size_t * table, int ** wog, int ** rwog, int *** Bm){
  /* ============== */
  /* Initialization */
  /* ============== */

  /* Calculate F() function for |A| = 0 */
  for(int g = 0; g < p; g++){
	int tid = omp_get_thread_num();
	int A; /* dummy */
	fdb[0][g][0].sc = Paraos_score(sb[tid], g, &A, 0);
	fdb[0][g][0].len = 0;
	fdb[0][g][0].B = 0; /* dummy */
#ifdef PARAOS_DEBUG
	fprintf(ytLib_ERR, "F(g=%d,A={}) = s(g=%d,A={}) = %f\n", g, g, fdb[0][g][0].sc);
#endif
  }/* for g < p */

  /* Calculate Q() for |A| = 0 */
  for(int j = 0; j < p; j++){
	qdb[0][p - j - 1].sc = fdb[0][j][0].sc;
	qdb[0][p - j - 1].pi[0] = j;
	qdb[0][p - j - 1].idx[0] = 0;
	qdb[0][p - j - 1].len[0] = 0;
#ifdef PARAOS_DEBUG
	fprintf(ytLib_ERR, "Q(M({%d})) = %f\n", j, qdb[0][p - j - 1].sc);
#endif
  }/* for j < p */


  /* ========= */
  /* Main Loop */
  /* ========= */

  /* |A| = a */
  for(int a = 1; a < p; a++){
	struct timeval start_time;
	struct timeval end_time;
	time_t current_time;
	//double sec;

	/* ------------------------------------- */
	/* Calculate F(g, A) from F(g, A - {a}). */
	/* ------------------------------------- */
	current_time = time(NULL);
	fprintf(ytLib_ERR, "\nStart Calculation for |A| = %d at %s", 
			a, ctime(&current_time));
	fflush(ytLib_ERR);
	gettimeofday(&start_time, NULL);

#ifdef _OPENMP
	os_calc_F_mp(sb, p, max_parents, a, fdb, table, wog, rwog);
#else
	os_calc_F(sb, p, max_parents, a, fdb, table, wog, rwog);
	//	os_calc_F(sf, sb, p, max_parents, a, fdb, table, wog, rwog, Bm);
#endif

	/* Calculate Q(A') where |A'| = |A| + 1 from results of
	   Q(A) and F(g,A) */
	os_calc_Q(a + 1, fdb, qdb, p, table, wog);
	//os_calc_Q(a + 1, fdb, qdb, p, table, wog, Bm);
	
	current_time = time(NULL);
	fprintf(ytLib_ERR, "Finished at %s", ctime(&current_time));
	gettimeofday(&end_time, NULL);
	//sec = Util_timeval_span(&start_time, &end_time);
	//fprintf(ytLib_ERR, "Split time: ");
	//Util_print_timeval_span(ytLib_ERR, &start_time, &end_time);
	fflush(ytLib_ERR);
  }/* for a */

  fprintf(ytLib_ERR, "os: Best sc = %f\n", qdb[(p + 1) % 2][0].sc);
  fprintf(ytLib_ERR, "X-BESTSCORE\t%f\n", qdb[(p + 1) % 2][0].sc);
  fprintf(ytLib_ERR, "os: Best pi : (");
  for(int i = 0; i < p; i++){
    if(i != 0) fprintf(ytLib_ERR, ",");
    fprintf(ytLib_ERR, "%d", qdb[(p + 1) % 2][0].pi[i]);
  }
  fprintf(ytLib_ERR, ")\n");
}
/* os_exec() =============================================================== */

/* F(g, A) function to calculate the best set of parents.
   sf, sb : score function and its initialized buffer.
   g : target gene ID
   A : array containing the parents
   a : size of parents

   fdb : array containing all scores of g with sets of parents whose
         size is |A| - 1.  The index needs to be for the combinations
		 without 'g'.

   p : number of total nodes.

   wog : full index table to sets witout 'g'. (wog: V => V - {g})
 */
FDB os_F(void * sb, int g, int * A, int a, FDB * fdb, int p,
		 int * wog, uint64_t * table, int max_parents){
		 //		 int * wog, size_t * table, int max_parents, int *** Bm){

  int B[p], BB[p]; /* B and BB requires only a - 1 elements. But here,
					  to avoid to allocate size 0, we fix the size to 'p'.
				      Note that 'a' can be 1 here, so a - 1 becomes 0. */
  FDB ret;
  size_t flag = (size_t) -1;
  int am = a - 1;
  float best_sc;

  
  /* Calculates the score for the full parent set of A,
	 That is. s(g,A). */
  if(a <= max_parents){
	best_sc = Paraos_score(sb, g, A, a);
#ifdef PARAOS_DEBUG
	fprintf(ytLib_ERR, "    s(g=%d, A={", g);
	Util_fprint_ar(ytLib_ERR, A, a);
	fprintf(ytLib_ERR, "}) = %f\n", best_sc);
	fflush(ytLib_ERR);
#endif
  } else {
	best_sc = 1000000000;
  }

  /* Preparation to generate the subsets of A. */
  //B[0] = -1;
  //int bm = 0;
  for(int i = 0; i < a - 1; i++) B[i] = i;

  /* ------------------------------------ */
  /* Calculates the parent index for fdb. */
  /* ------------------------------------ */
  /*  Searches the best scored pattern of A - {a} */
  for(int i = 0; i < a; i++){
	size_t rli;
	/* At first, obtain the subset of A, that is, A - {a}.  Here, 'B' is a 
	   combination vector of length a - 1 from a set {0, ..., a - 1} 
	   regardless of 'g'.  Thus,  you need to map 'B' to that of A - {g}.
	   A[B[j]] is a true index of the j-th element of 'B'.
	   By using 'wog', it converts a true index to that of V - {g}.
	*/
	//combo(a, am, B);
	for(int j = 0; j < am; j++) BB[j] = wog[A[B[j]]];
	//for(int j = 0; j < am; j++) BB[j] = wog[A[Bm[a][bm][j]]];
	//bm++;
	B[am - i - 1] = am - i;

	/* Then, obtain the reverse-lexicographic index (RLI) of 'BB'. */
	rli = ytCombo_index(p - 1, am, BB, table, p);
#ifdef PARAOS
	fprintf(ytLib_ERR, "    rli=%zu\n", rli);
	fprintf(ytLib_ERR, "    F(g=%d, A={", g);
	Util_fprint_ar2(ytLib_ERR, B, am, A);
	fprintf(ytLib_ERR, "}) (id=%zu) = %f\n", rli, fdb[rli].sc);
	fflush(ytLib_ERR);
#endif

	/* Compare the score */
	if(fdb[rli].sc < best_sc){
	  /* Update the best choise */
	  best_sc = fdb[rli].sc;
	  flag = rli;
	}
  }/* for i < a */
  
  if(flag != (size_t) -1){
	ret = fdb[flag];
  } else {
	/* A is the best combination */
	ret.sc = best_sc;
	ret.B = ytCombo_index(p, a, A, table, p);
	ret.len = a;
  }

  #ifdef PARAOS_DEBUG
  #pragma omp critical
  {
    fprintf(ytLib_ERR, "F(g=%d, A={", g);
    Util_fprint_ar(ytLib_ERR, A, a);
    fprintf(ytLib_ERR, "})=%f\n",best_sc);
    fflush(ytLib_ERR);
  }
  #endif
  #ifdef PARAOS_DEBUG
#pragma omp critical
  {
  fprintf(ytLib_ERR, "-- F(g=%d, A={", g);
  Util_fprint_ar(ytLib_ERR, A, a);
  fprintf(ytLib_ERR, "}) ------------------ \n");
  fprintf(ytLib_ERR, "B=%zu,len=%zu",(size_t) ret.B, (size_t) ret.len);
  fflush(ytLib_ERR);
  fprintf(ytLib_ERR, "    best = %f,  ", best_sc);
  fprintf(ytLib_ERR, "B={");
  //Util_fprint_id(ytLib_ERR, p, ret.len, ret.B, table, p);
  fprintf(ytLib_ERR, "}\n");
  fprintf(ytLib_ERR, "--------------------------------------\n");
  }
  #endif

  return ret;
}
/* os_F() ================================================================== */

void os_calc_F(void ** sb, int p, int max_parents, int a, 
			   FDB *** fdb, uint64_t * table, int ** wog, int ** rwog){
			   //FDB *** fdb, size_t * table, int ** wog, int ** rwog, int *** Bm){
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "!!! F() !!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  fprintf(ytLib_ERR, "  F(g, A)\n");
  fflush(ytLib_ERR);
#endif
  size_t num = ytCombo_choose(p - 1, a) - 1;
  int A[p], AA[p];
  
  /* Visit all nodes */
  for(int g = 0; g < p; g++){
	size_t id = num;
	//fprintf(ytLib_ERR, "DEBUG: id=%zu\n", id);

	/* Visit all parent combinations */
	AA[0] = -1;
	for(size_t i = 0; i <= num; i++){
	  ytCombo_generate(p - 1, a, AA);
	  
#ifdef PARAOS_DEBUG
	  fprintf(ytLib_ERR, "os: AA={");
	  Util_fprint_ar(ytLib_ERR, AA, i);
	  fprintf(ytLib_ERR, "}\n");
#endif
	  
	  for(int k = 0; k < a; k++) A[k] = rwog[g][AA[k]];
	  
#ifdef PARAOS_DEBUG
	  fprintf(ytLib_ERR, "os: A={");
	  Util_fprint_ar(ytLib_ERR, A, i);
	  fprintf(ytLib_ERR, "}\n");
#endif
	  
	  fdb[a % 2][g][id] = os_F(sb[0], g, A, a, fdb[(a + 1) % 2][g], 
							   //							   p, wog[g], table, max_parents, Bm);
							   p, wog[g], table, max_parents);
#ifdef PARAOS_DEBUG
	  fprintf(ytLib_ERR, "single: fdb[%d][%d][%zu].sc = %f\n", 
			  a % 2, g, id, fdb[a % 2][g][id].sc); fflush(ytLib_ERR);
	  ytCombo_combination(p, fdb[a % 2][g][id].len, fdb[a % 2][g][id].B, A, table, p);
	  fprintf(ytLib_ERR, "single:   B={");
	  Util_fprint_ar(ytLib_ERR, A, fdb[a % 2][g][id].len);
	  fprintf(ytLib_ERR, "}\n");
#endif
	  id--;
	}/* for i */
  }/* for g < p */
  
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"); fflush(ytLib_ERR);
#endif
}
/* os_calc_F() ============================================================= */

#ifdef _OPENMP
void os_calc_F_mp(void ** sb, const int p, int max_parents, int a, 
				  FDB *** fdb, uint64_t * table, int ** wog, int ** rwog){

  const size_t num = ytCombo_choose(p - 1, a) - 1;
  /* Visit all nodes */
  for(int g = 0; g < p; g++){
	fprintf(ytLib_ERR, "os: g=%d\n", g);

	/* Visit all parent combinations */
#ifdef _OPENMP
#pragma omp parallel for
#endif
	for(size_t i = 0; i <= num; i++){
	  int A[PARAOS_MAX_NODES];
	  int AA[PARAOS_MAX_NODES];
	  int tid = omp_get_thread_num();
	  size_t id = num - i;
	  ytCombo_combination(p - 1, a, id, AA, table, p);
	  
	  for(int k = 0; k < a; k++) A[k] = rwog[g][AA[k]];
	  
	  fdb[a % 2][g][id] = os_F(sb[tid], g, A, a, fdb[(a + 1) % 2][g], 
							   p, wog[g], table, max_parents);
#ifdef PARAOS_DEBUG
#ifdef _OPENMP
#pragma omp critical
	  {
#endif
		fprintf(ytLib_ERR, "os(%d):   AA={", tid);
		Util_fprint_ar(ytLib_ERR, AA, a);
		fprintf(ytLib_ERR, "}\n");
		fprintf(ytLib_ERR, "os(%d):   A={", tid);
		Util_fprint_ar(ytLib_ERR, A, a);
		fprintf(ytLib_ERR, "}\n");
		fprintf(ytLib_ERR, "os(%d):   fdb[%d][%d][%zu].sc = %f\n", 
				tid, a % 2, g, id, fdb[a % 2][g][id].sc); fflush(ytLib_ERR);
		ytCombo_combination(p, fdb[a % 2][g][id].len, fdb[a % 2][g][id].B, A, table, p);
		fprintf(ytLib_ERR, "os(%d):   B={", tid);
		Util_fprint_ar(ytLib_ERR, A, fdb[a % 2][g][id].len);
		fprintf(ytLib_ERR, "}\n");
#ifdef _OPENMP
	  }
#endif
#endif
	}/* for i */
  }/* for g < p */
  
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"); fflush(ytLib_ERR);
#endif
}
/* os_calc_F_mp() ========================================================== */
#endif /* ifdef _OPENMP */

/* Calc Q(A) for |A| = a */
//void os_calc_Q(int a, FDB *** fdb, QDB ** qdb, int p, size_t * table, int ** wog, int *** Bm){
void os_calc_Q(int a, FDB *** fdb, QDB ** qdb, int p, uint64_t * table, int ** wog){

  const size_t num = ytCombo_choose(p, a) - 1;
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "  Q(M(A)) num=%zu\n", num); fflush(ytLib_ERR);
#endif
#ifdef _OPENMP
#pragma omp parallel for
#endif
  for(size_t i = 0; i <= num; i++){
	int A[PARAOS_MAX_NODES];
	size_t rli = num - i;
	ytCombo_combination(p, a, rli, A, table, p);
	os_Q(A, a, fdb[(a - 1) % 2], qdb[a % 2], p, wog, table, p, 
		 &(qdb[(a - 1) % 2][rli]));
		 //		 &(qdb[(a - 1) % 2][rli]), Bm);
  }/* for() */
}
/* os_calc_Q() ============================================================= */

void os_Q(int * A, int a, FDB ** fdb, QDB * qdb, int n, int ** wog, 
		  uint64_t * table, int tn, QDB * ret){
		  //		  size_t * table, int tn, QDB * ret, int *** Bm){

  float best = 1000000000;
  int best_g = -1;
  int am = a - 1;
  int AA[am], AAA[am], B[am];
  size_t fidx, qidx, best_qidx = 0, best_fidx = 0;

#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "-- Q({");
  Util_fprint_ar(ytLib_ERR, A, a);
  fprintf(ytLib_ERR, "}) -------------------------- \n");
  fflush(ytLib_ERR);
#endif

  //AA[0] = -1;
  //int bm = 0;
  for(int i = 0; i < am; i++) AA[i] = i;
  
  /* A (|A| = a) からサイズ a - 1 のすべての組み合わせ AA を作る */
  for(int i = 0; i < a; i++){
	int g = A[a - i - 1]; /* AA に含まれないノード */
	double sc;

	//combo(a, am, AA);
	for(int j = 0; j < am; j++){
	  //int v = A[Bm[a][bm][j]];
	  int v = A[AA[j]];
	  B[j] = v;
	  AAA[j] = wog[g][v];
	}
	//bm++;
	AA[am - i - 1] = am - i;
	fidx = ytCombo_index(n - 1, am, AAA, table, tn);
	qidx = ytCombo_index(n, am, B, table, tn);
	
	sc = fdb[g][fidx].sc + qdb[qidx].sc;

#ifdef PARAOS_DEBUG
	fprintf(ytLib_ERR, "    (F(g=%d,A={", g);
	Util_fprint_ar(ytLib_ERR, B, am);
	fprintf(ytLib_ERR, "}) = %f) + (Q(M({", fdb[g][fidx].sc);
	Util_fprint_ar(ytLib_ERR, B, am);
	fprintf(ytLib_ERR, "})) = %f) = %f\n", qdb[qidx].sc, sc);
	fflush(ytLib_ERR);
#endif
	
	if(sc < best){
	  best = sc;
	  best_g = g;
	  best_qidx = qidx;
	  best_fidx = fidx;
	}
  }/* for i < a */
  ret->sc = best;
  #ifdef PARAOS_DEBUG
  #pragma omp critical
  {
    fprintf(ytLib_ERR, "Q({");
    Util_fprint_ar(ytLib_ERR, A, a);
    fprintf(ytLib_ERR, "})=%f\n",best);
    fflush(ytLib_ERR);
  }
  #endif  
  for(int i = 0; i < am; i++){
	ret->pi[i] = qdb[best_qidx].pi[i];
	ret->idx[i] = qdb[best_qidx].idx[i];
	ret->len[i] = qdb[best_qidx].len[i];
  }
  ret->pi[am] = best_g;
  ret->idx[am] = fdb[best_g][best_fidx].B;
  ret->len[am] = fdb[best_g][best_fidx].len;

  #ifdef PARAOS_DEBUG
  #pragma omp critical
  {
     fprintf(ytLib_ERR, "Q({");
    Util_fprint_ar(ytLib_ERR, A, a);
    fprintf(ytLib_ERR, "})\n");
    fflush(ytLib_ERR);
    fprintf(ytLib_ERR, "      Best = %f,  g = %d, ", ret->sc, best_g);
    fprintf(ytLib_ERR, " pi: (");
    Util_fprint_ar(ytLib_ERR, ret->pi, a);
    fprintf(ytLib_ERR, ")\n");
    fprintf(ytLib_ERR, "        Pats: ");
    for(int i = 0; i < a; i++){
      fprintf(ytLib_ERR, " %d: {", ret->pi[i]);
      //Util_fprint_id(ytLib_ERR, n, ret->len[i], ret->idx[i], table, n);
      fprintf(ytLib_ERR, "}  ");
    }
    fprintf(ytLib_ERR, "\n");
    fprintf(ytLib_ERR, "--------------------------------------\n");
    fflush(ytLib_ERR);
  }
  #endif
}
/* os_Q() ================================================================== */

/* Generate the graph structure from the results stored in QDB. */
ytPCGraph * os_final_graph(int p, QDB * qdb, uint64_t * table){
  int R[p], A[p];
  int j, jj;
  ytPCGraph * g;

  g = ytPCGraph_new(p);

  for(j = 0; j < p; j++) R[qdb->pi[j]] = j;

  for(j = 0; j < p; j++){
	size_t k;
	jj = R[j];
	fprintf(ytLib_ERR, "%d (%d): ", j, jj); fflush(ytLib_ERR);

	ytCombo_combination(p, qdb->len[jj], qdb->idx[jj], A, table, p);
	Util_fprint_ar(ytLib_ERR, A, qdb->len[jj]);
	fprintf(ytLib_ERR, "\n");

	for(k = 0; k < qdb->len[jj]; k++){
	  ytPCGraph_addEdge(g, A[k], j);
	}
  }

  return g;
}
/* os_final_graph() ======================================================== */

void os_memcheck(void * p, char * msg){
  if(p == NULL){
	fprintf(ytLib_ERR, "ERROR: Failed to allocate memory: %s\n", msg);
	fflush(ytLib_ERR);
	exit(1);
  }
}
/* os_memcheck() =========================================================== */

/* End of file. */
