/*
  Parallel OS Algorithm  Version 9 aka ParaOS-DC.
  Copyright (C) 2016-2017  Yoshinori Tamada  <tamada A T ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "paraos.h"
#include "Args.h"
#include "Util.h"
#include "score.h"

#include "ytLib.h"
#include "util/ytIntArray.h"
#include "util/ytDoubleArray.h"
#include "util/ytTimer.h"
#include "math/ytCombo.h"
#include "math/ytPCGraph.h"
#include "io/ytFile.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <float.h>
#include <mpi.h>
#include <assert.h>
#include <omp.h>

/*
  The structure storing F(v,A).  The optimal subset choice of A is stored as a 
  48 bit lexicographical index and stored in the fields 'Bl' and 'Bu'.
  
  The order of fileds is important.  They are listed so that the size is 12
  bytes in total.
 */
typedef struct {
  float sc;    /* score 4 bytes */
  uint32_t Bl; /* lower 4 bytes (32 bits) for B */
  uint16_t Bu; /* upper 2 bytes (16 bits) for B */
  uint8_t  v;  /* node   1 byte */
  uint8_t len; /* length of B: 1 byte */
} para9_F;

/*
  The structure for storing sub indices of F(v,A).
 */
typedef struct {
  uint64_t id;
  uint64_t subid;
} para9_Sub;

typedef struct {
  uint64_t id;
  uint64_t subid;
  uint8_t index;
  uint8_t v;
} para9_Qsub;

typedef struct {
  uint64_t * table; /* combination index table */
  int ** W;  /* W[i][j] represents the index of 'j' in V into V \ {'i'} */
  int ** R;  /* R[i][j] represents the index for 'j' in V \ {'i'} into V */

  para9_F ** F;
  para9_Sub * Fsub;
  uint64_t ** Fsubpos;
  /* Fsubpos[tid][r] represents the first Fsub index for Rank r. */

  uint8_t ** Q;
  para9_Qsub * Qsub;
  uint64_t ** Qsubpos;
  para9_Qsub * QFsub;
  uint64_t ** QFsubpos;

  int *** B; /* B[slot][i] represents a map for sub-combinations except for i. */

  /*
    Fflag, Qflag, and QFflag are used for indepenent sub index rank calculation
    by threads.
   */
  int ** Fflag;
  int ** Qflag;
  int ** QFflag;

  int * srcnums;
  int * srcnums2;
  int * srcs;
  size_t srcsSize;
  int * displs;

  ytIntArray * Ftable;
  ytIntArray * Qtable;

  ytIntArray * dstTableF;
  ytIntArray * dstTableQ;
  ytIntArray * Fdstsrc;
  ytIntArray * Qdstsrc;

  uint8_t * Qtmp;
  size_t Qtmp_elm_size;
  size_t QsendSize;
  size_t Qsize;
  size_t QMsize;

  ytTimer * M_timer;

  int * commSize;
  MPI_Comm * comm;
} para9_DB;

typedef struct {
  uint64_t total;
  uint64_t quot;
  uint64_t rem;
  uint64_t size;
} para9_Index;

typedef struct {
  int send;
  int recv;
  int calc;
  int next;
} para9_BuffIndex;

#define max(a,b) (a>b?a:b)
#define min(a,b) (a<b?a:b)

#include "sort.para9_Sub.c"
#include "sort.para9_Qsub.c"
#include "sort.table.c"

/*
  Calculates quot, rem, and size from 'total', and sets them in para9_Index.
  
  min : minimum jobs for a rank/thread.
 */
void para9_calcIndex(para9_Index * I, const uint64_t total, const int size, const uint64_t minElms){
  I->total = total;
  I->size = size;
  I->quot = total / size;

  if(I->quot < minElms){
    I->size = total / minElms;
    if(I->size == 0) I->size = 1;
    I->quot = total / I->size;
  }
  
  I->rem = total % I->size;
}/* para9_calcIndex() ====================================================== */

void para9_splitJobs(const para9_Index * I, const int rank,
                     uint64_t * bg, uint64_t * ed, uint64_t * elm){
  if(rank < (int) I->rem){
	*bg = (uint64_t) ((I->quot + 1) * rank);
	*ed = (uint64_t) (*bg + I->quot + 1);
	*elm = I->quot + 1;
  } else if(rank < I->size){
	*bg = (uint64_t) ((I->quot + 1) * I->rem + I->quot * (rank - I->rem));
	*ed = (uint64_t) (*bg + I->quot);
	*elm = I->quot;
  } else {
    *bg = 0;
    *ed = 0;
    *elm = 0;
  }
}/* para9_splitJobs() ====================================================== */

size_t para9_getFSize(){
  return sizeof(para9_F); /* It should be 12. */
}
size_t para9_getQSize(int a){
  return 4 + 8 * a;
}

/*!\brief Write a value of F(g,A) into the buffer.
 *\param buff    pointer to the buffer.
 *\param score   The score of F().
 *\param B       RLI on B - {g} representing the best parent combination.
 *\param len     Length of B.
 */
void para9_writeF(para9_F * buff, 
                  const uint8_t v, const uint64_t B, const uint8_t len,
                  const float score){
  buff->sc = score;
  buff->Bl = B & 0xffffffff;
  buff->Bu = B >> 32;
  buff->v = v;
  buff->len = len;
}/* para9_writeF() ========================================================== */

void para9_readF(const para9_F * buff,
                 uint8_t * v, uint64_t * B, uint8_t * len, float * score){
  *score = buff->sc;
  *B = buff->Bl;
  *B |= ((uint64_t) buff->Bu) << 32;
  *v = buff->v;
  *len = buff->len;
}/* para9_readF() =========================================================== */

void para9_memcpy4(void * dst, const void * src, size_t num){
  uint32_t * dst2 = (uint32_t *) dst;
  const uint32_t * src2 = (const uint32_t *) src;

  for(size_t i = 0; i < num; i++) *(dst2++) = *(src2++);
}/* para9_memcpy4() ======================================================== */

void para9_printTimers(int p, ytDoubleArray * timers){
  double total = 0.0;
  double maxTotal = 0.0;
  for(int i = 0; i < p; i++){
    double sum = 0.0;
    double max = -FLT_MAX;

    for(int j = 0; j < ytDoubleArray_size(timers + i); j++){
      double v = ytDoubleArray_get(timers + i, j);
      sum += v;
      if(max < v) max = v;
    }
    
    fprintf(ytLib_ERR, "%d: %f (", i + 1, sum);
    for(int j = 0; j < ytDoubleArray_size(timers + i); j++){
      double v = ytDoubleArray_get(timers + i, j);
      fprintf(ytLib_ERR, " %f", v);
    }
    fprintf(ytLib_ERR, " )\n");

    total += sum;
    maxTotal += max;
  }
  fprintf(ytLib_ERR, "TOTAL: %f\n", total);
  fprintf(ytLib_ERR, "MAX TOTAL: %f\n", maxTotal);
  fflush(ytLib_ERR);
}/* para9_printTimers() ==================================================== */

void para9_alloc_error(const char * str){
  fprintf(ytLib_ERR, "FATAL ERROR: memory allocation failed for %s\n", str);
  fflush(ytLib_ERR);
  exit(1);
}/* para9_alloc_error() ==================================================== */

size_t para9_alloc(int p, int procs, int rank, para9_DB * db, int threads, 
                   int dry, uint64_t minElms){
  size_t total = 0;
  size_t mem;

  fprintf(ytLib_ERR, "========================\n");
  fprintf(ytLib_ERR, "Checking type sizes...\n");
  fprintf(ytLib_ERR, "========================\n");
  fflush(ytLib_ERR);

  fprintf(ytLib_ERR, "sizeof(para9_F) = %zu\n", sizeof(para9_F));
  if(sizeof(para9_F) != 12){
    fprintf(ytLib_ERR, "FATAL ERROR: sizeof(para9_F) != 12\n");
    exit(1);
  }
  fflush(ytLib_ERR);
  fprintf(ytLib_ERR, "sizeof(para9_Qsub) = %zu\n", sizeof(para9_Qsub));

  fprintf(ytLib_ERR, "sizeof(float) = %zu\n", sizeof(float));
  if(sizeof(float) != 4){
    fprintf(ytLib_ERR, "FATAL ERROR: sizeof(float) != 4\n");
    exit(1);
  }

  fprintf(ytLib_ERR, "int max = ");
  ytFile_printComma_size_t(ytLib_ERR, INT32_MAX);
  fprintf(ytLib_ERR, "\n");

  fprintf(ytLib_ERR, "\n");
  fprintf(ytLib_ERR, "=====================\n");
  fprintf(ytLib_ERR, "Allocating memory...\n");
  fprintf(ytLib_ERR, "=====================\n");
  fprintf(ytLib_ERR, "minElms=%llu\n", minElms);
  fflush(ytLib_ERR);

  /* Index table */
  mem = sizeof(uint64_t) * p * p; total += mem;
  db->table = malloc(mem);
  ytCombo_indexTable(p, db->table);
  fprintf(ytLib_ERR, "table: ");
  ytFile_printComma_size_t(ytLib_ERR, mem);
  fprintf(ytLib_ERR, " [B]\n");
  fflush(ytLib_ERR);
  if(db->table == NULL) para9_alloc_error("table");

  /* W[i][j] represents the index of 'j' in V into V \ {'i'} */
  mem = sizeof(int *) * p; total += mem;
  if(!dry) db->W = malloc(mem);
  mem = sizeof(int) * p * p; total += mem;
  if(!dry) db->W[0] = malloc(mem);

  /* R[i][j] represents the index for 'j' in V \ {'i'} into V */
  mem = sizeof(int *) * p; total += mem;
  if(!dry) db->R = malloc(mem);
  mem = sizeof(int) * p * p; total += mem;
  if(!dry) db->R[0] = malloc(mem);

  if(!dry){
    for(int i = 0; i < p; i++){
      db->W[i] = db->W[0] + (i * p);
      db->R[i] = db->R[0] + (i * p);
      int v = 0;
      for(int j = 0; j < p; j++){
        if(i == j){
          db->W[i][j] = -1;
        } else {
          db->W[i][j] = v;
          db->R[i][v] = j;
          v++;
        }
      }
      db->R[i][p - 1] = -1;
    }
  }

  /* comm */
  mem = sizeof(MPI_Comm) * p; total += mem;
  db->comm = malloc(mem);
  mem = sizeof(int) * p; total += mem;
  db->commSize = malloc(mem);

  /* Q & F ----------------------------------------------------------------- */
  /* For |A| = a < p */
  
  size_t Q_max = 0, F_max = 0, F_total_max = 0;
  size_t Fsub_max = 0, Qsub_max = 0;
  size_t Qtmp_max = 0;
  uint64_t prevQsize, prevFsize = 1;
  para9_Index F, Q;
  F.size = 0;
  Q.size = 0;
  for(int a = 0; a <= p; a++){
    prevQsize = Q.size;
    para9_calcIndex(&Q, ytCombo_chooseTable(p, a, db->table, p), procs, minElms);
    size_t Q_elm = Q.quot;
    if(Q.rem != 0) Q_elm++;
    size_t Q_mem = Q_elm * para9_getQSize(a) + 4; /* +4 = +1 four-byte integer 
                                                      for the source rank. */
    if(Q_max < Q_mem) Q_max = Q_mem;

    size_t m = Q_elm * a * max(para9_getQSize(a), para9_getFSize());
    if(Qtmp_max < m) Qtmp_max = m;

#ifdef PARAOS_DEBUG
    fprintf(ytLib_ERR, "a=%d; Q.total=%llu, Q.size=%llu, Q_elm=%zu, Q.size=%llu, prevQsize=%llu\n", 
            a, Q.total, Q.size, Q_elm, Q.size, prevQsize);
    fflush(ytLib_ERR);
#endif

    if(a < p){
      prevFsize = F.size;
      para9_calcIndex(&F, ytCombo_chooseTable(p - 1, a, db->table, p) * p, procs, minElms);
      size_t F_elm = F.quot;
      if(F.rem != 0) F_elm++;
      //F_elm += 1; /* +1 for +1 four-byte integer for the source rank */
      if(F_max < F_elm){
        F_max = F_elm;
        F_total_max = F.total;
      }

      size_t v = F_elm * a;
      if(Fsub_max < v) Fsub_max = v;
      v = Q_elm * a;
      if(Qsub_max < v) Qsub_max = v;

#ifdef PARAOS_DEBUG
      fprintf(ytLib_ERR, "a=%d; F.total=%llu, F_elm=%zu, F.size=%llu, prevFsize=%llu\n", a, F.total, F_elm, F.size, prevFsize);
      fflush(ytLib_ERR);
#endif
    }
    if(0 < a){
      int maxSize = max(max(F.size,prevFsize), max(Q.size, prevQsize));
#ifdef PARAOS_DEBUG
      fprintf(ytLib_ERR, "a=%d, maxSize=%d\n", a, maxSize);
      fflush(ytLib_ERR);
#endif
      if(!dry){
        db->commSize[a - 1] = maxSize;
        if(maxSize == procs){
          db->comm[a - 1] = MPI_COMM_WORLD;
        } else {
          MPI_Comm_split(MPI_COMM_WORLD, rank < maxSize ? 0 : MPI_UNDEFINED, rank, &db->comm[a - 1]);
        }
      }
    }
  }
  fprintf(ytLib_ERR, "INT_MAX=");
  ytFile_printComma_size_t(ytLib_ERR, INT_MAX);
  fprintf(ytLib_ERR, "\n");  
  fprintf(ytLib_ERR, "F_total_max=");
  ytFile_printComma_size_t(ytLib_ERR, F_total_max);
  fprintf(ytLib_ERR, "\n");
  fprintf(ytLib_ERR, "F_max (elms)=");
  ytFile_printComma_size_t(ytLib_ERR, F_max);
  fprintf(ytLib_ERR, "\n");
  fprintf(ytLib_ERR, "F_max (memory)=");
  ytFile_printComma_size_t(ytLib_ERR, F_max * sizeof(para9_F) + 4);
  fprintf(ytLib_ERR, "\n");
  fprintf(ytLib_ERR, "F_max (send elms)=");
  ytFile_printComma_size_t(ytLib_ERR, (F_max * sizeof(para9_F) + 4) / 4);
  fprintf(ytLib_ERR, "\n");
  fprintf(ytLib_ERR, "Fsub_max=");
  ytFile_printComma_size_t(ytLib_ERR, Fsub_max);
  fprintf(ytLib_ERR, "\n");
  fprintf(ytLib_ERR, "Qsub_max=");
  ytFile_printComma_size_t(ytLib_ERR, Qsub_max);
  fprintf(ytLib_ERR, "\n");
  fflush(ytLib_ERR);

  int Fslots = 4;
  int Qslots = 3;
  mem = sizeof(para9_F **) * Fslots; total += mem;
  if(!dry){
    db->F = malloc(mem);
  }
  mem = (F_max + 1) * sizeof(para9_F) * Fslots; total += mem;
  fprintf(ytLib_ERR, "F[]: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);
  if(!dry){
    db->F[0] = malloc(mem);
    for(int i = 1; i < Fslots; i++){
      db->F[i] = db->F[i - 1] + (F_max + 1);
    }
  }

  mem = sizeof(uint8_t **) * Qslots; total += mem;
  if(!dry){
    db->Q = malloc(mem);
  }
  mem = Q_max * Qslots; total += mem; /* Q_max is in byte */
  fprintf(ytLib_ERR, "Q[]: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);
  if(!dry){
    db->Q[0] = malloc(mem);
    for(int i = 1; i < Qslots; i++){
      db->Q[i] = db->Q[i - 1] + Q_max;
    }
  }

  /* Fsub */
  mem = sizeof(para9_Sub) * Fsub_max; total += mem;
  if(!dry){
    db->Fsub = malloc(mem);
  }
  fprintf(ytLib_ERR, "Fsub: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);

  mem = sizeof(uint64_t *) * threads; total += mem;
  if(!dry) db->Fsubpos = malloc(mem);
  mem = sizeof(uint64_t) * threads * (procs + 1); total += mem;
  if(!dry){
    db->Fsubpos[0] = malloc(mem);
    for(int tid = 1; tid < threads; tid++){
      db->Fsubpos[tid] = db->Fsubpos[tid - 1] + (procs + 1);
    }
  }

  /* Qsub */
  mem = sizeof(para9_Qsub) * Qsub_max; total += mem;
  if(!dry) db->Qsub = malloc(mem);
  fprintf(ytLib_ERR, "Qsub: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);

  mem = sizeof(uint64_t *) * threads; total += mem;
  if(!dry) db->Qsubpos = malloc(mem);
  mem = sizeof(uint64_t) * threads * (procs + 1); total += mem;
  if(!dry){
    db->Qsubpos[0] = malloc(mem);
    for(int tid = 1; tid < threads; tid++){
      db->Qsubpos[tid] = db->Qsubpos[tid - 1] + (procs + 1);
    }
  }

  /* QFsub */
  mem = sizeof(para9_Qsub) * Qsub_max; total += mem;
  if(!dry) db->QFsub = malloc(mem);
  fprintf(ytLib_ERR, "QFsub: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);
  if(!dry && db->QFsub == NULL) para9_alloc_error("QFsub");

  mem = sizeof(uint64_t *) * threads; total += mem;
  if(!dry) db->QFsubpos = malloc(mem);
  mem = sizeof(uint64_t) * threads * (procs + 1); total += mem;
  if(!dry){
    db->QFsubpos[0] = malloc(mem);
    for(int tid = 1; tid < threads; tid++){
      db->QFsubpos[tid] = db->QFsubpos[tid - 1] + (procs + 1);
    }
  }

  /* Qtmp */
  mem = Qtmp_max; total += mem;
  if(!dry) db->Qtmp = malloc(mem);
  fprintf(ytLib_ERR, "Qtmp[]: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);
  if(!dry && db->Qtmp == NULL) para9_alloc_error("Qtmp[]");

  /* B --------------------------------------------------------------------- */
  size_t subtotal = 0;
  mem = sizeof(int **) * 2; subtotal += mem;
  if(!dry) db->B = malloc(mem);
  mem = sizeof(int *) * p * 2; subtotal += mem;
  if(!dry){
    db->B[0] = malloc(mem);
    db->B[1] = db->B[0] + p;
  }
  mem = sizeof(int) * p * p * 2; subtotal += mem;
  if(!dry){
    int * ptr = malloc(mem);
    for(int j = 0; j < 2; j++){
      for(int i = 0; i < p; i++){
        db->B[j][i] = ptr; ptr += p;
      }
    }
  }
  fprintf(ytLib_ERR, "B[][]: ");
  ytFile_printComma_size_t(ytLib_ERR, subtotal);
  fprintf(ytLib_ERR, " [B]\n");
  total += subtotal;

  /* Fflag, Qflag, QFflag -------------------------------------------------- */
  mem = sizeof(int *) * threads; total += mem;
  if(!dry) db->Fflag = malloc(mem);
  mem = sizeof(int) * threads * procs; total += mem;
  if(!dry){
    db->Fflag[0] = malloc(mem);
    for(int tid = 1; tid < threads; tid++){
      db->Fflag[tid] = db->Fflag[tid - 1] + procs;
    }
  }
  fprintf(ytLib_ERR, "Fflag[]: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  mem = sizeof(int *) * threads; total += mem;
  if(!dry) db->Qflag = malloc(mem);
  mem = sizeof(int) * threads * procs; total += mem;
  if(!dry){
    db->Qflag[0] = malloc(mem);
    for(int tid = 1; tid < threads; tid++){
      db->Qflag[tid] = db->Qflag[tid - 1] + procs;
    }
  }
  fprintf(ytLib_ERR, "Qflag[]: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  mem = sizeof(int *) * threads; total += mem;
  if(!dry) db->QFflag = malloc(mem);
  mem = sizeof(int) * threads * procs; total += mem;
  if(!dry){
    db->QFflag[0] = malloc(mem);
    for(int tid = 1; tid < threads; tid++){
      db->QFflag[tid] = db->QFflag[tid - 1] + procs;
    }
  }
  fprintf(ytLib_ERR, "QFflag[]: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");

  /* srcnums, srcnum2, srcs */
  mem = sizeof(int) * procs * 2; total += mem;
  if(!dry) db->srcnums = malloc(mem);
  fprintf(ytLib_ERR, "srcnums: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");

  mem = sizeof(int) * procs; total += mem;
  if(!dry) db->srcnums2 = malloc(mem);
  fprintf(ytLib_ERR, "srcnums2: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");

  db->srcs = NULL;
  db->srcsSize = 0;
  
  mem = sizeof(int) * procs; total += mem;
  if(!dry){
    db->displs = malloc(mem);
    db->displs[0] = 0;
  }
  fprintf(ytLib_ERR, "displs: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");

  db->Ftable = ytIntArray_arrayNew(procs);
  mem += ytIntArray_arrayMemSize(db->Ftable, procs);
  total += mem;
  fprintf(ytLib_ERR, "Ftable: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  
  db->Qtable = ytIntArray_arrayNew(procs);
  mem += ytIntArray_arrayMemSize(db->Qtable, procs);
  total += mem;
  fprintf(ytLib_ERR, "Qtable: ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");

  /* dstTableF */ 
  //mem = sizeof(int *) * procs; total += mem;
  //db->dstTableF = malloc(mem);
  //fprintf(ytLib_ERR, "dstTableF: ");
  //ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  //fprintf(ytLib_ERR, " [MB]\n");
  //mem = sizeof(int) * procs * (procs * 2); total += mem;
  //db->dstTableF[0] = malloc(mem);
  //for(int i = 1; i < procs; i++){
  //db->dstTableF[i] = db->dstTableF[i - 1] + (procs * 2);
  //}
  //fprintf(ytLib_ERR, "dstTableF[]: ");
  //ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  //fprintf(ytLib_ERR, " [MB]\n");
  //fflush(ytLib_ERR);
  db->dstTableF = ytIntArray_arrayNew(procs);
  total += ytIntArray_arrayMemSize(db->dstTableF, procs);

  /* dstTableQ */
  //mem = sizeof(int *) * procs; total += mem;
  //db->dstTableQ = malloc(mem); 
  //fprintf(ytLib_ERR, "dstTableQ: ");
  //ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  //fprintf(ytLib_ERR, " [MB]\n");
  //mem = sizeof(int) * procs * (procs * 2); total += mem;
  //db->dstTableQ[0] = malloc(mem);
  //for(int i = 1; i < procs; i++){
  //db->dstTableQ[i] = db->dstTableQ[i - 1] + (procs * 2);
  //}
  //fprintf(ytLib_ERR, "dstTableQ[]: ");
  //ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  //fprintf(ytLib_ERR, " [MB]\n");
  //fflush(ytLib_ERR);
  db->dstTableQ = ytIntArray_arrayNew(procs);
  total += ytIntArray_arrayMemSize(db->dstTableQ, procs);  
  
  /* Fdstsrc, Qdstsrc */
  db->Fdstsrc = ytIntArray_arrayNew(procs);
  total += ytIntArray_arrayMemSize(db->Fdstsrc, procs);
  db->Qdstsrc = ytIntArray_arrayNew(procs);
  total += ytIntArray_arrayMemSize(db->Qdstsrc, procs);
  
  return total;
}/* para9_alloc() ========================================================== */


/*\brief Calculates F sub indices for this rank and thread.
 *
 * The calculated indices are set in db->Fsub[i] where (FT_bg * a) <= i < (FT_ed * a),
 * then sorted in ascending order of F sub indices.
 *
 * This is called inside the OpenMP parallel regsion.
 *
 * F sub indices are the indices of F values calculated at the previous step.
 *
 * Output: db->Fsub, db->Fsubpos, and db->Fflag.
 *
 *\param p  the total number of vertices.
 *\param a  the current size of A to calculate F(v,A)s.
 *\param FMS  choose(p - 1, a - 1).
 *\param FS  choose(p - 1, a), i.e., the number of F(v,A)s where v is fixed.
 *\param F_bg  the first index of F(v,A) being calculated in this rank.
 *\param FT_bg  the offset from F_bg for this thread.
 */
void para9_prepare_F(const int p, const int a, const int MPI_size,
                     const para9_Index * FM, const uint64_t FMS, const uint64_t FS,
                     const uint64_t F_bg,
                     const uint64_t FT_bg, const uint64_t FT_ed, const uint64_t FT_elm,
                     para9_DB * db, const int tid, const int threads){
  /*
    Traverse F(v,A) for (F_bg + FT_bg) <= index(F(v,A)) < (F_ed + FT_ed).
   */
  for(uint64_t ii = FT_bg, offset = FT_bg * a; ii < FT_ed; ii++, offset += a){
    int A[PARAOS_MAX_NODES];
    int v = (ii + F_bg) / FS;
    int id = (ii + F_bg) % FS;
    
    ytCombo_combination2(p - 1, a, id, A, db->R[v], db->table, p);
#ifdef PARAOS_DUMP
#pragma omp critical
    {
      fprintf(ytLib_ERR, "{%d} For F(%d,", tid, v);
      ytCombo_print(ytLib_ERR, a, A);
      fprintf(ytLib_ERR, "):%llu\n", ii + F_bg);
      fflush(ytLib_ERR);
    }
#endif
    /* Traverse sub indices of F(v,A). That is, F(v,A\{j}) where j represents the j-th 
       element of A. */
    for(int j = 0; j < a; j++){
      uint64_t subid = v * FMS + ytCombo_index3(p - 1, a - 1, db->B[a % 2][j], A, db->W[v],
                                                db->table, p);
      db->Fsub[offset + j].id = ii + F_bg; /* index of F(v,A) */
      db->Fsub[offset + j].subid = subid;
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        //fprintf(ytLib_ERR, "{%d}   B[%d][%d]=", tid, slotA, j);
        //ytCombo_print2(ytLib_ERR, a - 1, db->B[slotA][j], A);
        //fprintf(ytLib_ERR, "\n");
        fprintf(ytLib_ERR, "{%d}   Fsub[%llu]=F(%d,", tid, offset + j, v);
        ytCombo_print2(ytLib_ERR, a - 1, db->B[a % 2][j], A);
        fprintf(ytLib_ERR, "):%llu={id=%llu, subid=%llu}\n", 
                subid, ii + F_bg, subid);
        fflush(ytLib_ERR);
      }
#endif
    }/* for j < a */
  }/* for ii < FT_ed */

  /* Sorts sub IDs. */
  ytSort_exec_para9_Sub(db->Fsub + FT_bg * a, 
                        db->Fsub + (FT_bg + FT_elm) * a);
#ifdef PARAOS_DUMP
#pragma omp critical
  {
    for(uint64_t i = FT_bg * a; i < (FT_bg + FT_elm) * a; i++){
      fprintf(ytLib_ERR, "{%d} Fsub[%llu]={id=%llu, subid=%llu}\n",
              tid, i, db->Fsub[i].id, db->Fsub[i].subid);
    }
    fflush(ytLib_ERR);
  }
#endif
  
  /* Set db->Fsubpos[t][r] that represents the sorted Fsub[] position corresponding to 
     rank r where t is the thread id. */
  int prevRank = -1;
  uint64_t prevId = -1;
  uint64_t num = 0;
  for(uint64_t ii = FT_bg * a; ii < (FT_bg + FT_elm) * a; ii++){
    uint64_t id = db->Fsub[ii].subid;
    if(id != prevId){
      int rank = Util_splitJobsRank(FM->quot, FM->rem, id);
      db->Fflag[tid][rank] = 1;
      prevId = id;
      if(rank != prevRank){
        for(int r = prevRank + 1; r < rank; r++){
          db->Fsubpos[tid][r] = ii;
        }
        db->Fsubpos[tid][rank] = ii;
        prevRank = rank;
      }
    }
    num++;
  }/* for ii < FT_elm * a */
  for(int r = prevRank + 1; r <= MPI_size; r++){
    db->Fsubpos[tid][r] = (FT_bg + FT_elm) * a;
  }
#ifdef PARAOS_DUMP
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} Fsubpos --- \n", tid);
    for(int r = 0; r < MPI_size; r++){
      fprintf(ytLib_ERR, "   rank=%d; %llu\n", 
              r, db->Fsubpos[tid][r]);
    }
    fflush(ytLib_ERR);
  }
#endif
}/* para9_prepare_F() ====================================================== */

/*
  Q_bg : The first Q index for this rank.
  QT_bg : The first Q offset index for this thread. The Q index for this thread
  can be calculated as Q_bg + QT_bg
*/
void para9_prepare_Q(const int p, const int a, const int MPI_size,
                     const para9_Index * QM, const para9_Index * FM, 
                     const uint64_t FMS,
                     const uint64_t Q_bg,
                     const uint64_t QT_bg, const uint64_t QT_ed, const uint64_t QT_elm,
                     para9_DB * db, const int tid, const int threads){
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "{%d} prepare_Q(); \n",
          tid);
  
#endif
  for(uint64_t ii = QT_bg, offset = QT_bg * a; ii < QT_ed; ii++, offset += a){
    int A[PARAOS_MAX_NODES];
    ytCombo_combination(p, a, ii + Q_bg, A, db->table, p);
#ifdef PARAOS_DUMP
#pragma omp critical
    {
      fprintf(ytLib_ERR, "{%d} For Q(", tid);
      ytCombo_print(ytLib_ERR, a, A);
      fprintf(ytLib_ERR, "):%llu\n", ii + Q_bg);
    }
#endif
    for(int j = 0; j < a; j++){
      uint64_t Qsubid = ytCombo_index2(p, a - 1, db->B[a % 2][j], 
                                       A, db->table, p);
      int v = A[j];
      uint64_t Fsubid = v * FMS +
        ytCombo_index3(p - 1, a - 1, db->B[a % 2][j], A, db->W[v], db->table, p);

      db->Qsub[offset + j].id = ii + Q_bg;
      db->Qsub[offset + j].subid = Qsubid;
      db->Qsub[offset + j].index = j;

      db->QFsub[offset + j].id = ii + Q_bg;
      db->QFsub[offset + j].subid = Fsubid;
	  db->QFsub[offset + j].index = j;
      
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        fprintf(ytLib_ERR, "{%d}   Qsub[%llu]=Q(", tid, offset + j);
        ytCombo_print2(ytLib_ERR, a - 1, db->B[a % 2][j], A);
        fprintf(ytLib_ERR, "):%llu, index=%d\n", Qsubid, j);
        fprintf(ytLib_ERR, "{%d}   QFsub[%llu]=F(%d,", tid, offset + j, v);
        ytCombo_print2(ytLib_ERR, a - 1, db->B[a % 2][j], A);
        fprintf(ytLib_ERR, "):%llu, index=%d\n", Fsubid, j);
      }
#endif
    }/* for j < a */
  }/* for ii < QT_ed */
  
#ifdef PARAOS_DUMPX
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} Qsub <BEFORE SORT>\n", tid);
    for(uint64_t i = QT_bg * a; i < (QT_bg + QT_elm) * a; i++){
      fprintf(ytLib_ERR, "{%d} Qsub[%llu]={id=%llu, subid=%llu, index=%d}\n",
              tid, i, db->Qsub[i].id, db->Qsub[i].subid, db->Qsub[i].index);
    }
    fflush(ytLib_ERR);

    fprintf(ytLib_ERR, "{%d} QFsub <BEFORE SORT>\n", tid);
    for(uint64_t i = QT_bg * a; i < (QT_bg + QT_elm) * a; i++){
      fprintf(ytLib_ERR, "{%d} QFsub[%llu]={id=%llu, subid=%llu, index=%d}\n",
              tid, i, db->QFsub[i].id, db->QFsub[i].subid, db->QFsub[i].index);
    }
    fflush(ytLib_ERR);
  }
#endif

  /* Q sub combinations ---------------------------------------------------- */

  ytSort_exec_para9_Qsub(db->Qsub + QT_bg * a,
                         db->Qsub + (QT_bg + QT_elm) * a);
#ifdef PARAOS_DUMP
#pragma omp critical
  {
    for(uint64_t i = QT_bg * a; i < (QT_bg + QT_elm) * a; i++){
      fprintf(ytLib_ERR, "{%d} Qsub[%llu]={id=%llu, subid=%llu, index=%d}\n",
              tid, i, db->Qsub[i].id, db->Qsub[i].subid, db->Qsub[i].index);
    }
    fflush(ytLib_ERR);
  }
#endif

  for(int i = 0; i < MPI_size; i++) db->Qflag[tid][i] = 0;
  int prevRank = -1;
  uint64_t prevId = -1;
  if(a > 1){
    for(uint64_t ii = QT_bg * a; ii < (QT_bg + QT_elm) * a; ii++){
      uint64_t id = db->Qsub[ii].subid;
      if(id != prevId){
        int rank = Util_splitJobsRank(QM->quot, QM->rem, id);
        db->Qflag[tid][rank] = 1;
        prevId = id;
        if(rank != prevRank){
          for(int r = prevRank + 1; r < rank; r++){
            db->Qsubpos[tid][r] = ii;
          }
          db->Qsubpos[tid][rank] = ii;
          prevRank = rank;
        }
      }
    }/* for ii < (QT_bg + QT_elm) * a */
  }/* if a > 1. */
  for(int r = prevRank + 1; r <= MPI_size; r++){
    db->Qsubpos[tid][r] = (QT_bg + QT_elm) * a; /* Fill the terminal mark */
  }
#ifdef PARAOS_DUMP
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} Qsubpos ------ \n", tid);
    for(int r = 0; r < MPI_size; r++){
      fprintf(ytLib_ERR, "  rank=%d; %llu\n",
              r, db->Qsubpos[tid][r]);
    }
    fflush(ytLib_ERR);
  }
#endif

  /* QF sub combinations --------------------------------------------------- */

#ifdef PARAOS_DUMPX
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} QFsub <BEFORE SORT>\n", tid);
    for(uint64_t i = QT_bg * a; i < (QT_bg + QT_elm) * a; i++){
      fprintf(ytLib_ERR, "{%d} QFsub[%llu]={id=%llu, subid=%llu}\n",
              tid, i, db->QFsub[i].id, db->QFsub[i].subid);
    }
    fflush(ytLib_ERR);
  }
#endif

  ytSort_exec_para9_Qsub(db->QFsub + QT_bg * a,
                         db->QFsub + (QT_bg + QT_elm) * a);
  
#ifdef PARAOS_DUMP
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} QFsub <AFTER SORT>\n", tid);
    for(uint64_t i = QT_bg * a; i < (QT_bg + QT_elm) * a; i++){
      fprintf(ytLib_ERR, "{%d} QFsub[%llu]={id=%llu, subid=%llu, index=%d}\n",
              tid, i, db->QFsub[i].id, db->QFsub[i].subid, db->QFsub[i].index);
    }
    fflush(ytLib_ERR);
  }
#endif

  for(int i = 0; i < MPI_size; i++) db->QFflag[tid][i] = 0;

  prevRank = -1;
  prevId = -1;
  for(uint64_t ii = QT_bg * a; ii < (QT_bg + QT_elm) * a; ii++){
    uint64_t id = db->QFsub[ii].subid;
    if(id != prevId){
      int rank = Util_splitJobsRank(FM->quot, FM->rem, id);
      db->QFflag[tid][rank] = 1;
      prevId = id;
      if(rank != prevRank){
        for(int r = prevRank + 1; r < rank; r++){
          db->QFsubpos[tid][r] = ii;
        }
        db->QFsubpos[tid][rank] = ii;
        prevRank = rank;
      }
    }
  }/* for ii < (QT_bg + QT_elm) * a */
  for(int r = prevRank + 1; r <= MPI_size; r++){
    db->QFsubpos[tid][r] = (QT_bg + QT_elm) * a;
  }
#ifdef PARAOS_DUMP
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} QFsubpos ------ \n", tid);
    for(int r = 0; r < MPI_size; r++){
      fprintf(ytLib_ERR, "  rank=%d; %llu\n",
              r, db->QFsubpos[tid][r]);
    }
    fflush(ytLib_ERR);
  }
#endif

}/* para9_prepare_Q() ====================================================== */

void para9_calc_F(const int rank, const int recv, const int calc,
                  const uint64_t F_bg, const uint64_t FMR_bg, 
#ifdef PARAOS_DUMP
                  int p, int a, uint64_t FS, uint64_t FMS, 
#endif
                  para9_DB * db, const int tid){
  uint64_t Fsub_bg = db->Fsubpos[tid][rank];
  uint64_t Fsub_ed = db->Fsubpos[tid][rank + 1];

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "<%d> para9_calc_F(); rank=%d, Fsub_bg=%llu, Fsub_ed=%llu\n", 
            tid, rank, Fsub_bg, Fsub_ed);
  }
#endif

  uint64_t FsubidPrev = -1;
  float sc = 0.0;
  for(uint64_t i = Fsub_bg; i < Fsub_ed; i++){
    uint64_t Fsubid = db->Fsub[i].subid;
    if(Fsubid != FsubidPrev){
      sc = db->F[recv][Fsubid - FMR_bg].sc;
      FsubidPrev = Fsubid;
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int B[PARAOS_MAX_NODES];
        int v = Fsubid / FMS;
        uint64_t id = Fsubid % FMS;
        ytCombo_combination2(p - 1, a - 1, id, B, db->R[v], db->table, p);
        fprintf(ytLib_ERR, "<%d> Received F(%d,", tid, v);
        ytCombo_print(ytLib_ERR, a - 1, B);
        fprintf(ytLib_ERR, "):%llu=F[%d][%llu]=%f\n", Fsubid, recv, Fsubid - FMR_bg, sc);
        fflush(ytLib_ERR);
      }
#endif
    }

    uint64_t Fid = db->Fsub[i].id; /* Target F index */
    float old = db->F[calc][Fid - F_bg].sc;

    if(sc < old){
      db->F[calc][Fid - F_bg] = db->F[recv][Fsubid - FMR_bg];
	}

#ifdef PARAOS_DUMP
#pragma omp critical
	{
	  int v = Fid / FS;
	  uint64_t id = Fid % FS;
	  int B[PARAOS_MAX_NODES];
	  ytCombo_combination2(p - 1, a, id, B, db->R[v], db->table, p);
	  fprintf(ytLib_ERR, "<%d> F[%d][%llu]=F(%d,",
			  tid, calc, Fid - F_bg, v);
	  ytCombo_print(ytLib_ERR, a, B);
	  fprintf(ytLib_ERR, "):%llu <- old:%f  VS  new:%f\n", Fid, old, sc);
	  fflush(ytLib_ERR);
	}
#endif
  }/* for i where Fsub_bg <= i < Fsub_ed */
}/* para9_calc_F() ========================================================= */

void para9_makeQ(uint8_t * buff,
                 const uint8_t * Qsrc, const para9_F * Fsrc,
                 int a, float newsc){
  uint32_t * buff4 = (uint32_t *) buff;
  size_t s = 2 * (a - 1);

  /* First, copy F at the last. */
  *(buff4 + 1 + s) = *(((const uint32_t *) Fsrc) + 1);
  *(buff4 + 2 + s) = *(((const uint32_t *) Fsrc) + 2);
  
  /* Score */
  *((float *) buff4++) = newsc; /* 4 byte copy */

  /* B, v & len for (a - 1) nodes */
  para9_memcpy4(buff4, ((uint32_t *) Qsrc) + 1, s);
}/* para9_makeQ() =========================================================== */

void para9_calc_Q(const int rank, const int recv, const int calc,
                  const uint64_t Q_bg, const uint64_t QMR_bg, const int a,
#ifdef PARAOS_DUMP
                  int p,
#endif
                  para9_DB * db, const int tid){

  uint64_t Qsub_bg = db->Qsubpos[tid][rank];
  uint64_t Qsub_ed = db->Qsubpos[tid][rank + 1];
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "<%d> para9_calc_Q(); rank=%d, Qsub_bg=%llu, Qsub_ed=%llu\n",
            tid, rank, Qsub_bg, Qsub_ed);
    fflush(ytLib_ERR);
  }
#endif
  int QsubidPrev = -1;
  for(uint64_t i = Qsub_bg; i < Qsub_ed; i++){
    uint64_t Qid = db->Qsub[i].id;
    uint64_t Qsubid = db->Qsub[i].subid;
    uint8_t offset = db->Qsub[i].index;
    size_t addr = ((Qid - Q_bg) * a + offset) * db->Qtmp_elm_size; 
    float FMsc = *((float *) (db->Qtmp + addr));

    if(QsubidPrev != Qsubid){
      QsubidPrev = Qsubid;

#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int A[PARAOS_MAX_NODES];
        ytCombo_combination(p, a - 1, Qsubid, A, db->table, p);
        fprintf(ytLib_ERR, "<%d> Received Q(", tid);
        ytCombo_print(ytLib_ERR, a - 1, A);
        fprintf(ytLib_ERR, "):%llu=Q[%d][%llu]=%f\n", Qsubid, recv, Qsubid - QMR_bg,
				*((float *) (db->Q[recv] + (Qsubid - QMR_bg) * db->QMsize)));
		fflush(ytLib_ERR);
      }
#endif
    }
    
    if(FMsc == 0.0){
      /* F not available. Just writes it to Qtmp. */
      para9_memcpy4(db->Qtmp + addr,
                    db->Q[recv] + (Qsubid - QMR_bg) * db->QMsize,
                    2 * (a - 1) + 1);
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int A[PARAOS_MAX_NODES];
        ytCombo_combination(p, a, Qid, A, db->table, p);
        fprintf(ytLib_ERR, "<%d> Q[%d][%llu]=Q(", tid, calc, Qid - Q_bg);
        ytCombo_print(ytLib_ERR, a, A);
        fprintf(ytLib_ERR, "); Qtmp[%llu:%d] <- Q(", Qid - Q_bg, offset);
        fprintf(ytLib_ERR, "):%llu=%f\n", i, *((float *) (db->Q[recv] + (Qsubid - QMR_bg) * db->QMsize)));
        fflush(ytLib_ERR);
      }
#endif
    } else {
      /* F available */
      float QMsc = *((float *) (db->Q[recv] + (Qsubid - QMR_bg) * db->QMsize));
      float old = *((float *) (db->Q[calc] + (Qid - Q_bg) * db->Qsize));
      float newSc = FMsc + QMsc;
      if(newSc < old){
        para9_makeQ(db->Q[calc] + (Qid - Q_bg) * db->Qsize,
                    db->Q[recv] + (Qsubid - QMR_bg) * db->QMsize,
                    (para9_F *) (db->Qtmp + addr), a, newSc);
      }
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int A[PARAOS_MAX_NODES];
        ytCombo_combination(p, a, Qid, A, db->table, p);
        fprintf(ytLib_ERR, "<%d> Q[%d][%llu] = Q(", tid, calc, Qid - Q_bg);
        ytCombo_print(ytLib_ERR, a, A);
        fprintf(ytLib_ERR, "):%llu <- old:%f VS new:%f=(Q:%f+F:%f)\n", Qid, 
                (old == FLT_MAX ? 9999.9 : old), newSc, QMsc, FMsc);
        fflush(ytLib_ERR);
      }
#endif
    }
  }/* for i where Qsub_bg <= i < Qsub_ed */
}/* para9_calc_Q() ========================================================= */

void para9_calc_QF1(const int rank, const int recv, const int calc, 
                    const uint64_t Q_bg, const uint64_t FMR_bg, const uint64_t FMS,
                    int p, para9_DB * db, const int tid){
  uint64_t QFsub_bg = db->QFsubpos[tid][rank];
  uint64_t QFsub_ed = db->QFsubpos[tid][rank + 1];
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "<%d> para9_calc_QF1(); rank=%d, QFsub_bg=%llu, QFsub_ed=%llu\n",
            tid, rank, QFsub_bg, QFsub_ed);
    fflush(ytLib_ERR);
  }
#endif
  uint64_t FsubidPrev = -1;
  float FMsc = 0.0;
  for(uint64_t i = QFsub_bg; i < QFsub_ed; i++){
    uint64_t Fsubid = db->QFsub[i].subid;
    if(Fsubid != FsubidPrev){
      FMsc = db->F[recv][Fsubid - FMR_bg].sc;
      FsubidPrev = Fsubid;

#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int v = Fsubid / FMS;
        fprintf(ytLib_ERR, "<%d> Received F(%d,{}):%llu=%f\n",
                tid, v, Fsubid, FMsc);
        fflush(ytLib_ERR);
      }
#endif
    }

    uint64_t Qid = db->QFsub[i].id;
    
    para9_makeQ(db->Q[calc] + (Qid - Q_bg) * db->Qsize, NULL,
                &(db->F[recv][Fsubid - FMR_bg]), 1, FMsc);

#ifdef PARAOS_DUMP
#pragma omp critical
    {
      int v = Fsubid / FMS;
      fprintf(ytLib_ERR, "<%d> Q[%d][%llu] <- Q({})+F(%d,{}) = %f\n",
              tid, calc, Qid - Q_bg, v, FMsc);
      fflush(ytLib_ERR);
    }
#endif
    
  }/* for i  */
}/* para9_calc_QF1() ======================================================= */

void para9_calc_QF(const int p, const int a,
                   const int rank, const int recv, const int calc,
                   const uint64_t Q_bg, const uint64_t FMR_bg,
                   const uint64_t FMS, para9_DB * db, const int tid){
  if(a == 1){
    para9_calc_QF1(rank, recv, calc, Q_bg, FMR_bg, FMS, p, db, tid);
    return;
  }
  
  uint64_t QFsub_bg = db->QFsubpos[tid][rank];
  uint64_t QFsub_ed = db->QFsubpos[tid][rank + 1];
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "<%d> para9_calc_QF(); rank=%d, QFsub_bg=%llu, QFsub_ed=%llu\n",
            tid, rank, QFsub_bg, QFsub_ed);
    fflush(ytLib_ERR);
  }
#endif
  uint64_t FsubidPrev = -1;
  float FMsc = 0.0;
  for(uint64_t i = QFsub_bg; i < QFsub_ed; i++){
    uint64_t Fsubid = db->QFsub[i].subid;
    if(Fsubid != FsubidPrev){
      FMsc = db->F[recv][Fsubid - FMR_bg].sc;
      FsubidPrev = Fsubid;
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int B[PARAOS_MAX_NODES];
        int v = Fsubid / FMS;
        uint64_t id = Fsubid % FMS;
        ytCombo_combination2(p - 1, a - 1, id, B, db->R[v], db->table, p);
        fprintf(ytLib_ERR, "{%d} Received F(%d,", tid, v);
        ytCombo_print(ytLib_ERR, a - 1, B);
        fprintf(ytLib_ERR, "):%llu=%f\n", Fsubid, FMsc);
        fflush(ytLib_ERR);
      }
#endif
    }

    uint64_t Qid = db->QFsub[i].id;
    uint8_t offset = db->QFsub[i].index;
    size_t addr = ((Qid - Q_bg) * a + offset) * db->Qtmp_elm_size;
    float QMsc = *((float *) (db->Qtmp + addr));
    //fprintf(ytLib_ERR, "{%d} Qtmp[%llu:%d]=%f\n", tid, Qid - Q_bg, offset, QMsc);
    fflush(ytLib_ERR);
    if(QMsc == 0.0){
      /* Q not available.  Just writes the received F sub element into Qtmp. */
      para9_memcpy4(db->Qtmp + addr,
                    (uint8_t *) &(db->F[recv][Fsubid - FMR_bg]), 3);
#ifdef PARAOS_DUMP
#pragma omp critical
      {
        int A[PARAOS_MAX_NODES];
        ytCombo_combination(p, a, Qid, A, db->table, p);
        fprintf(ytLib_ERR, "{%d} Q[%d][%llu]=Q(", tid, calc, Qid - Q_bg);
        ytCombo_print(ytLib_ERR, a, A);
        fprintf(ytLib_ERR, "):%llu; Qtmp[%llu:%d] <- F():%llu = %f\n",
                Qid, Qid - Q_bg, offset, Fsubid, FMsc);
        fflush(ytLib_ERR);
      }
#endif
      
    } else {
      /* Q available */
      float old = *((float *) (db->Q[calc] + (Qid - Q_bg) * db->Qsize));
      float newSc = QMsc + FMsc;
      if(newSc < old){
        para9_makeQ(db->Q[calc] + (Qid - Q_bg) * db->Qsize,
                    db->Qtmp + addr,
                    &(db->F[recv][Fsubid - FMR_bg]), a, newSc);
      }
#ifdef PARAOS_DUMP
#pragma omp critical
      fprintf(ytLib_ERR, "<%d> Q[%d][%llu] <- old:%f VS new:%f=(Q:%f+F:%f)\n",
              tid, calc, Qid - Q_bg, (old == FLT_MAX ? 9999.9 : old), 
              newSc, QMsc, FMsc);
      fflush(ytLib_ERR);
#endif
    }
  }/* for i where QFsub_bg <= i < QFsub_ed */
}/* para9_calc_QF() ======================================================== */

/* Calculates the source nodes of this rank.

   Required source ranks are already calculated in para9_prepareF() and 
   para9_prepareQ().  The results are stored in Fflag, QFflag and Qflag.
   These source ranks are calculated seperately in threads. Fflag[t][r] == 1 
   if rank r is required where t is the thread.  
   
   The calculated source ranks are set in Fsrc and Qsrc.
 */
void para9_calcSrc(const int MPI_rank, const int MPI_size, const int threads,
                   int * const * Fflag, int * const * Qflag, int * const * QFflag,
                   ytIntArray * Fsrc, ytIntArray * Qsrc,
                   int * Fprevsrc, int * Qprevsrc){
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "calcSrc(); MPI_rank=%d, MPI_size=%d, threads=%d\n", MPI_rank, MPI_size, threads);
    fflush(ytLib_ERR);
  }
#endif

  ytIntArray_clear(Fsrc);
  ytIntArray_clear(Qsrc);

  /* Visit all the ranks */
  for(int i = 0; i < MPI_size; i++){
    int flagF = Fflag[0][i];
    int tid = 1;
    /* Here we just need to know a thread that requires data in rank i exsits. */
    while(flagF == 0 && tid < threads){
      flagF = Fflag[tid++][i];
    }

    int flagQF = QFflag[0][i];
    tid = 1;
    while(flagQF == 0 && tid < threads){
      flagQF = QFflag[tid++][i];
    }
    int value = -1;
    if(flagF && flagQF){
      value = 0x3000000 | i;
    } else if(flagF){
      value = 0x1000000 | i;
    } else if(flagQF){
      value = 0x2000000 | i;
    }
    if(i == MPI_rank){
      *Fprevsrc = value;
    } else if(value != -1){
      ytIntArray_add(Fsrc, value);
    }

    int flagQ = Qflag[0][i];
    tid = 1;
    while(flagQ == 0 && tid < threads){
      flagQ = Qflag[tid++][i];
    }
    if(flagQ){
      if(i == MPI_rank){
        *Qprevsrc = MPI_rank;
      } else {
        ytIntArray_add(Qsrc, i);
      }
    }
  }/* for i < MPI_size */

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "F sources (%zu):", ytIntArray_size(Fsrc));
    for(int i = 0; i < ytIntArray_size(Fsrc); i++){
      fprintf(ytLib_ERR, " %d(%d)", 
              ytIntArray_get(Fsrc, i) & 0xffffff,
              ytIntArray_get(Fsrc, i) >> 24);
    }
    fprintf(ytLib_ERR, "\n");
    fprintf(ytLib_ERR, "Q sources (%zu):", ytIntArray_size(Qsrc));
    for(int i = 0; i < ytIntArray_size(Qsrc); i++){
      fprintf(ytLib_ERR, " %d", ytIntArray_get(Qsrc, i));
    }
    fprintf(ytLib_ERR, "\n");
    fflush(ytLib_ERR);
    if(Fflag[MPI_rank] || QFflag[MPI_rank]){
      fprintf(ytLib_ERR, "  Local data processing required for F.\n");
      fflush(ytLib_ERR);
    }
    if(Qflag[MPI_rank]){
      fprintf(ytLib_ERR, "  Local data processing required for Q.\n");
      fflush(ytLib_ERR);
    }
  }
#endif
}/* para9_calcSrc() ======================================================== */

typedef struct para9_int_pair {
  int a;
  int b;
} para9_int_pair_t;

int para9_int_pair_cmp(const void * xx, const void * yy){
  const para9_int_pair_t * x = (const para9_int_pair_t *) xx;
  const para9_int_pair_t * y = (const para9_int_pair_t *) yy;
  if(x->b > y->b){
    return -1;
  } else if(y->b > x->b){
    return 1;
  } else {
    return 0;
  }
}

#include "sort.para9_int_pair_t.c"

/*
 *\param table  input distination lists where 'table + j' contains the list of 
 *             the j-th rank.
 *\param T
 */
void para9_optimize(int procs, ytIntArray * table, ytIntArray * T, ytIntArray * result){
  /* Sort based on the number of destinations */
  para9_int_pair_t pairs[procs];
  for(int i = 0; i < procs; i++){
    pairs[i].a = i;
    pairs[i].b = ytIntArray_size(table + i); 
  }
  //qsort(pairs, procs, sizeof(para9_int_pair_t), para9_int_pair_cmp);
  ytSort_exec_para9_int_pair_t(pairs, pairs + procs);

  /* Set the longest paths into the first row of T. */
  int i = pairs[0].a;
  int max = pairs[0].b;
  for(int j = 0; j < max; j++){
    ytIntArray_set(T + 0, j, ytIntArray_get(table + i, j));
  }
  ytIntArray_set(T + 0, max, -2); /* termination marker */

  /* Set paths in T[ii][j] */
  int max_j = 0;
  int dst = 0; /* values to be set in T */
  int dst2 = 0; /* actual destination rank except high-bit flags */
  for(int ii = 1; ii < procs; ii++){
    int i = pairs[ii].a; /* target rank */
    int j = 0; /* column position of T[ii][j] */
    int jj = 0; /* position of dst for i */
    int prevjj = -1;
    while(jj < pairs[ii].b){
      /* If 'jj' is incremented, read the new destination */
      if(prevjj != jj){
        dst = ytIntArray_get(table + i, jj);
        dst2 = dst & 0xffffff;
      }
      int flag = 0;
      /* Checks if the dst exists already in the same column position. */
      for(int k = 0; k < ii; k++){
        if(ytIntArray_get(T + k, j) == -2){
          ytIntArray_set(T + k, j, -1);
          ytIntArray_set(T + k, j + 1, -2);
        } else if((ytIntArray_get(T + k, j) & 0xffffff) == dst2){
          /* The same destination found.  Fills -1 and moves to the next
             column. */
          ytIntArray_set(T + ii, j, -1);
          j++;       
          flag = 1;
          break;
        }
      }
      prevjj = jj;
      if(flag == 0){
        /* No overlap found.  Go next. */
        ytIntArray_set(T + ii, j, dst);
        jj++;
        j++;
      }
    }/* while() */

    if(max_j < j) max_j = j;
    for(; j <= max_j; j++){
      ytIntArray_set(T + ii, j, -2); /* fill -2 in the blanks */
    }
  }/* for ii < procs */

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "Before filling\n");
    for(int i = 0; i < procs; i++){
      int j = 0;
      fprintf(ytLib_ERR, "%d:", pairs[i].a);
      while(ytIntArray_get(T + i, j) != -2){
        fprintf(ytLib_ERR, " %2d", 
                ytIntArray_get(T + i, j) == -1 ?
                -1 : ytIntArray_get(T + i, j) & 0xffffff);
        j++;
      }
      fprintf(ytLib_ERR, "\n");
    }
    fflush(ytLib_ERR);
  }
#endif

  /* filling */
  int last[procs];
  int flag[procs]; /* Whether the number appears in the column */
  for(int i = 0; i < procs; i++) last[i] = -2;
  for(int j = max_j - 1; j >= 0; j--){
    memset(flag, 0, sizeof(int) * procs);
    for(int i = 0; i < procs; i++){
      if(ytIntArray_get(T + i, j) >= 0) flag[ytIntArray_get(T + i, j) & 0xffffff] = 1;
    }
    for(int i = 0; i < procs; i++){
      if(ytIntArray_get(T + i, j) == -1){
        if(last[i] == -2){
          ytIntArray_set(T + i, j, -2);
        } else {
          /* find the empty proc */
          for(int k = 0; k < procs; k++){
            int idx = (last[i] & 0xffffff) - k;
            if(idx < 0) idx += procs;
            if(!flag[idx]){
              ytIntArray_set(T + i, j, 0x4000000 | idx);
              flag[idx] = 1;
              break;
            }
          }
          if(ytIntArray_get(T + i, j) == -1){
            fprintf(ytLib_ERR, "FATAL ERROR: optimize()\n");
            fflush(ytLib_ERR);
            exit(1);
          }
        }
      }
      last[i] = ytIntArray_get(T + i, j);
    }/* for i < procs */
  }

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "Optimization result:\n");
    for(int i = 0; i < procs; i++){
      fprintf(ytLib_ERR, "%3d:", i);
      fprintf(ytLib_ERR, " %3d", pairs[i].a);
      int j = 0;
      while(ytIntArray_get(T + i, j) != -2){
        if(ytIntArray_get(T + i, j) == -1){
          fprintf(ytLib_ERR, "   - (-)");
        } else {
          fprintf(ytLib_ERR, " %3d(%d)",
                  ytIntArray_get(T + i, j) & 0xffffff,
                  ytIntArray_get(T + i, j) >> 24);
        }
        j++;
      }
      fprintf(ytLib_ERR, "\n");
    }
    fflush(ytLib_ERR);
  }
#endif

  /* format of 'result': dst1 src1 dst2 src2 ....  */

  for(int i = 0; i < procs; i++){
    ytIntArray_clear(result + i);
  }
  
  for(int i = 0; i < procs; i++){
    int j = 0;
    int dst = ytIntArray_get(T + i, j);
    int src = pairs[i].a | (dst & 0xff000000);
    while(dst != -2){
      int src2 = src & 0xffffff;
      int dst2 = dst & 0xffffff;
      while(ytIntArray_size(result + src2) < j * 2){
        ytIntArray_add(result + src2, -1);
      }
      ytIntArray_set(result + src2, j * 2, dst2);

      while(ytIntArray_size(result + dst2) < j * 2 + 1){
        ytIntArray_add(result + dst2, -1);
      }
      ytIntArray_set(result + dst2, j * 2 + 1, src2 | (dst & 0xff000000));

      src = dst;
      dst = ytIntArray_get(T + i, ++j);
    }
  }
  for(int i = 0; i < procs; i++){
    if(ytIntArray_size(result + i) % 2 == 1){
      ytIntArray_add(result + i, -1);
    }
  }

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "Optimized Flow:\n");
    for(int i = 0; i < procs; i++){
      fprintf(ytLib_ERR, "%3d:", i);
      for(int j = 0; j < ytIntArray_size(result + i); j++){
        int v = ytIntArray_get(result + i, j);
        fprintf(ytLib_ERR, " %d(%d)", 
                (v == -1 ? -1 : v & 0xffffff), v >> 24);
      }
      fprintf(ytLib_ERR, " (#. %zu)\n", ytIntArray_size(result + i));
    }
    fflush(ytLib_ERR);
  }
#endif
  
}/* para9_optimize() ======================================================= */

void para9_optimizeGather(const int MPI_rank, const int MPI_size,
                          ytIntArray * Fsrc, ytIntArray * Qsrc,
                          para9_DB * db, int a){
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "optimizeGather(); \n");
    fflush(ytLib_ERR);
  }
#endif
  
  int srcnum[2];
  srcnum[0] = (int) ytIntArray_size(Fsrc);
  srcnum[1] = (int) ytIntArray_size(Qsrc);

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "optimizeGather(); srcnum[0]=%d, srcnum[1]=%d\n", srcnum[0], srcnum[1]);
    fflush(ytLib_ERR);
  }
#endif
  /* At first, gather the numbers of the source ranks of all the ranks in db->srcnums. */
  MPI_Gather(srcnum, 2, MPI_INT, db->srcnums, 2, MPI_INT, 0, db->comm[a - 1]);
  ytTimer_sector(db->M_timer); /* 1 (Gather) */

#ifdef PARAOS_GTIME
  /* GTIME ID=2 */
          fprintf(ytLib_ERR, "X-GTIME:\t%d\t%d\t%d\t%.6f\n", 
                  MPI_rank, a, 2, ytTimer_getTime());
          fflush(ytLib_ERR);
#endif
  
  if(MPI_rank == 0){
    /* REMOVE IFDEF HERE FOR OBTAINING THE NUMBERS OF TARGET NODES
       USED IN THE PAPER. */
#ifdef PARAOS_DEBUG
#pragma omp critical
    {
      fprintf(ytLib_ERR, "Gathered source rank numbers.\n");
      int fmax = 0, qmax = 0;
      for(int r = 0; r < MPI_size; r++){
        fprintf(ytLib_ERR, "%d : {%d, %d}\n", 
                r, db->srcnums[r * 2], db->srcnums[r * 2 + 1]);
        if(fmax < db->srcnums[r * 2]) fmax = db->srcnums[r * 2];
        if(qmax < db->srcnums[r * 2 + 1]) qmax = db->srcnums[r * 2 + 1];
      }
      fprintf(ytLib_ERR, "FMax = %d\n", fmax);
      fprintf(ytLib_ERR, "QMax = %d\n", qmax);
      fflush(ytLib_ERR);
    }
#endif /* END OF THE BLOCK */
    db->srcnums2[0] = db->srcnums[0] + db->srcnums[1];
    for(int i = 1; i < MPI_size; i++){
      db->srcnums2[i] = db->srcnums[i * 2] + db->srcnums[i * 2 + 1];
      db->displs[i] = db->displs[i - 1] + db->srcnums2[i - 1];
    }
    int total = db->displs[MPI_size - 1] + db->srcnums2[MPI_size - 1];
    if(db->srcsSize < total){
      db->srcs = realloc(db->srcs, sizeof(int) * total);
    }
  }/* if(MPI_rank == 0) */
#ifdef PARAOS_DEBUG
  else {
#pragma omp critical
    {
      fprintf(ytLib_ERR, "optimizeGather(): MPI_Gather() finished.\n");
      fflush(ytLib_ERR);
    }
  }
#endif
  
  /* Add Q sources at the end of Fsrc to send them together. */
  ytIntArray_addIntArray(Fsrc, Qsrc);

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "optimizeGather(): |Fsrc|=%zu.\n", ytIntArray_size(Fsrc));
    fflush(ytLib_ERR);
  }
#endif
  /* Gathers F and Q source ranks in db->srcs in rank 0. */
  ytTimer_sector(db->M_timer); /* 2 */
  MPI_Gatherv(Fsrc->ptr, Fsrc->size, MPI_INT,
              db->srcs, db->srcnums2, db->displs, MPI_INT,
              0, db->comm[a - 1]);
  ytTimer_sector(db->M_timer); /* 3 (Gatherv) */

  /* Here, the received data (source rank lists) are stored in 'srcnums2'. */
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    if(MPI_rank == 0){
      fprintf(ytLib_ERR, "Received Source Ranks -----\n");
      fprintf(ytLib_ERR, " (Rank [FQtotal]: ranks...)\n");
      for(int r = 0; r < MPI_size; r++){
        fprintf(ytLib_ERR, "%d [%d]: ", r, db->srcnums2[r]);
        for(int i = 0; i < db->srcnums2[r]; i++){
          fprintf(ytLib_ERR, " %d", (db->srcs[db->displs[r] + i]) & 0xffffff);
        }
        fprintf(ytLib_ERR, "\n");
      }
      fprintf(ytLib_ERR, "---------------------------\n");
      fflush(ytLib_ERR);
    } else {
      fprintf(ytLib_ERR, "optimizeGather(): MPI_Gatherv() finished\n");
      fflush(ytLib_ERR);
    }
  }
#endif
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    if(MPI_rank == 0){
      int flag[MPI_size];
      int max = 0;
      fprintf(ytLib_ERR, "The number of non-redundant source ranks -----\n");
      fprintf(ytLib_ERR, " (Rank: Num)\n");
      for(int r = 0; r < MPI_size; r++){
        memset(flag, 0, sizeof(int) * MPI_size);
        for(int s = 0; s < db->srcnums2[r]; s++){
          flag[db->srcs[db->displs[r] + s] & 0xffffff] = 1;
        }
        int sum = 0;
        for(int s = 0; s < MPI_size; s++){
          if(flag[s]) sum++;
        }
        if(max < sum) max = sum;
        fprintf(ytLib_ERR, "%d\t%d\n", r, sum);
      }/* for r < MPI_size */
      fprintf(ytLib_ERR, "Max = %d\n", max);
    }/* if(MPI_rank == 0) */
  }
#endif
}/* para9_optimizeGather() ================================================= */

void para9_optimizerF(const int MPI_rank, const int MPI_size,
                     ytIntArray * Fsrc, para9_DB * db, int a){
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "para9_optimizerF()\n");
    fflush(ytLib_ERR);
  }
#endif
  
  /* Convert the source list to the destination list */
  for(int r = 0; r < MPI_size; r++){
    ytIntArray_clear(db->Ftable + r);
  }
  /* F */
  for(int r = 0; r < MPI_size; r++){
    for(int i = 0; i < db->srcnums[r * 2 + 0]; i++){
      int value = db->srcs[db->displs[r] + db->srcnums[r * 2] * 0 + i];
      int src = value & 0xffffff;
      int flag = value & 0xff000000;
      int dst = r;
      if(dst < src) dst += MPI_size;
      ytIntArray_add(db->Ftable + src, dst | flag);
    }
  }
  for(int r = 0; r < MPI_size; r++){
    ytSort_exec_int((db->Ftable + r)->ptr, (db->Ftable + r)->ptr + ytIntArray_size(db->Ftable + r));
    for(int i = 0; i < ytIntArray_size(db->Ftable + r); i++){
      int v = ytIntArray_get(db->Ftable + r, i) & 0xffffff;
      if(v >= MPI_size){
        int flag = ytIntArray_get(db->Ftable + r, i) & 0xff000000;
        ytIntArray_set(db->Ftable + r, i, (v - MPI_size) | flag);
      }
    }
  } 
    
#ifdef PARAOS_DEBUG
    fprintf(ytLib_ERR, "Ftable (destinations): Rank (type) ...\n");
    for(int r = 0; r < MPI_size; r++){
      fprintf(ytLib_ERR, "%d:", r);
      for(int i = 0; i < ytIntArray_size(db->Ftable + r); i++){
        fprintf(ytLib_ERR, " %d(%d)", 
                ytIntArray_get(db->Ftable + r, i) & 0xffffff,
                ytIntArray_get(db->Ftable + r, i) >> 24);
      }
      fprintf(ytLib_ERR, "\n");
    }
    fflush(ytLib_ERR);
#endif

    para9_optimize(MPI_size, db->Ftable, db->dstTableF, db->Fdstsrc);

}/* para9_optimizerF() ===================================================== */

void para9_optimizerQ(const int MPI_rank, const int MPI_size,
                      ytIntArray * Qsrc, para9_DB * db, int a){

#ifdef PAROS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "para9_optimizerQ()\n");
    fflush(ytLib_ERR);
  }
#endif

  /* Convert the source list to the destination list */
  for(int r = 0; r < MPI_size; r++){
    ytIntArray_clear(db->Qtable + r);
  }
  /* Q */
  for(int r = 0; r < MPI_size; r++){
    for(int i = 0; i < db->srcnums[r * 2 + 1]; i++){
      int value = db->srcs[db->displs[r] + db->srcnums[r * 2] * 1 + i];
      int src = value & 0xffffff;
      int flag = value & 0xff000000;
      int dst = r;
      if(dst < src) dst += MPI_size;
      ytIntArray_add(db->Qtable + src, dst | flag);
    }
  }
  for(int r = 0; r < MPI_size; r++){
    ytSort_exec_int((db->Qtable + r)->ptr, (db->Qtable + r)->ptr + ytIntArray_size(db->Qtable + r));
    for(int i = 0; i < ytIntArray_size(db->Qtable + r); i++){
      int v = ytIntArray_get(db->Qtable + r, i) & 0xffffff;
      if(v >= MPI_size){
        int flag = ytIntArray_get(db->Qtable + r, i) & 0xff000000;
        ytIntArray_set(db->Qtable + r, i, (v - MPI_size) | flag);
      }
    }
  } 
    
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "Qtable:\n");
  for(int r = 0; r < MPI_size; r++){
    fprintf(ytLib_ERR, "%d:", r);
    for(int i = 0; i < ytIntArray_size(db->Qtable + r); i++){
      fprintf(ytLib_ERR, " %d", ytIntArray_get(db->Qtable + r, i));
    }
    fprintf(ytLib_ERR, "\n");
  }
  fflush(ytLib_ERR);
#endif
  para9_optimize(MPI_size, db->Qtable, db->dstTableQ, db->Qdstsrc);
  
}/* para9_optimizerQ() ==================================================== */

/*
  Called from para9_taskOptimizerB().
 */
void para9_optimizer(const int MPI_rank, const int MPI_size,
                     ytIntArray * Fsrc, ytIntArray * Qsrc,
                     para9_DB * db, int a){

  if(MPI_rank == 0){
    /* Reuse srcnum */
    for(int i = 0; i < MPI_size; i++){
      db->srcnums[i * 2] = (int) ytIntArray_size(db->Fdstsrc + i);
      db->srcnums[i * 2 + 1] = (int) ytIntArray_size(db->Qdstsrc + i);
      db->srcnums2[i] = db->srcnums[i * 2] + db->srcnums[i * 2 + 1];
      if(i != 0){
        db->displs[i] = db->displs[i - 1] + db->srcnums2[i - 1];
      }
      if(i != 0){
        ytIntArray_addIntArray(db->Fdstsrc + 0, db->Fdstsrc + i);
      }
      ytIntArray_addIntArray(db->Fdstsrc + 0, db->Qdstsrc + i);
    }
  } /* if(MPI_rank == 0) */
  
  int srcnum[2];
  ytTimer_sector(db->M_timer); /* 6 */
  MPI_Scatter(db->srcnums, 2, MPI_INT, srcnum, 2, MPI_INT, 0, db->comm[a - 1]);
  ytTimer_sector(db->M_timer); /* 7 */
  
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "optimizer(): MPI_Scatter() finished; srcnum[0]=%d, srcnum[1]=%d\n",
            srcnum[0], srcnum[1]);
    fflush(ytLib_ERR);
  }
#endif
  
  if(ytIntArray_buffSize(Fsrc) < (srcnum[0] + srcnum[1])){
    ytIntArray_setBuffSize(Fsrc, srcnum[0] + srcnum[1]);
  }

  ytTimer_sector(db->M_timer); /* 8 */
  MPI_Scatterv(db->Fdstsrc->ptr, db->srcnums2, db->displs, MPI_INT,
               Fsrc->ptr, srcnum[0] + srcnum[1], MPI_INT, 0, db->comm[a - 1]);
  ytTimer_sector(db->M_timer); /* 9 */
  
  ytIntArray_setSize(Qsrc, srcnum[1]);
  ytIntArray_copy(Qsrc, 0, Fsrc->ptr + srcnum[0], srcnum[1]);
  ytIntArray_setSize(Fsrc, srcnum[0]);

#ifdef PAROS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "F flow (%zu):", ytIntArray_size(Fsrc) / 2);
    fflush(ytLib_ERR);
    for(int i = 0; i < srcnum[0]; i += 2){
      int dst = ytIntArray_get(Fsrc, i);
      int src = ytIntArray_get(Fsrc, i + 1);
      fprintf(ytLib_ERR, " %d:D%d,S%d(%d);", i / 2, 
              dst,
              (src == -1 ? -1 : src & 0xffffff), src >> 24);
      fflush(ytLib_ERR);
    }
    fprintf(ytLib_ERR, "\n");
    fprintf(ytLib_ERR, "Q flow (%zu):", ytIntArray_size(Qsrc) / 2);
    for(int i = 0; i < srcnum[1]; i += 2){
      int dst = ytIntArray_get(Qsrc, i);
      int src = ytIntArray_get(Qsrc, i + 1);
      fprintf(ytLib_ERR, " %d:D%d,S%d(%d);", i / 2, dst, 
              (src == -1 ? -1 : src & 0xffffff), src >> 24);
    }
    fprintf(ytLib_ERR, "\n");
    fflush(ytLib_ERR);
  }
#endif
  ytTimer_stopLap(db->M_timer); /* 10 last */

}/* para9_optimizer() ====================================================== */

typedef struct {
  int Fprevsrc;
  int Qprevsrc;

  uint64_t max_F;
  uint64_t max_Q;
  
  int Fi;
  int Qi;

  int dstsrc[4];

  int FsendSize;
  int QsendSize;

  ytIntArray * Fdst;
  ytIntArray * Qdst;

  MPI_Request reqs[4];

  MPI_Request testReqs[4];
  int testIDs[4];
  int testNum;

  ytTimer * C_timer;
  ytDoubleArray * C_timers;
  ytTimer * D_timer;
  ytDoubleArray * D_timers;
} para9_ThreadData;

/*

 */
typedef struct {
  int status;
  uint64_t scorePos;
  uint64_t scoreTotal;
  int p;
  int a;
  int threads;
  float g;
  int MPI_rank;
  int MPI_size;
  int assignedThreads;
  int finishedThreads;
  para9_ThreadData * td;
  para9_BuffIndex * Fb;
  para9_BuffIndex * Qb;
#ifdef PARAOS_DEBUG
  para9_DB * db;
#endif
} para9_TaskQueue;

void para9_initTaskQueue(para9_TaskQueue * q, int p, int a, int threads,
                         uint64_t totalScores, float g, para9_ThreadData * td, 
                         para9_BuffIndex * Fb, para9_BuffIndex * Qb,
                         int MPI_rank, int MPI_size){
  q->status = -1;
  q->scorePos = 0;
  q->scoreTotal = totalScores;
  q->p = p;
  q->a = a;
  q->threads = threads;
  q->assignedThreads = threads;
  q->finishedThreads = 0;
  q->g = g;
  q->td = td;
  q->Fb = Fb;
  q->Qb = Qb;
  q->MPI_rank = MPI_rank;
  q->MPI_size = MPI_size;
}/* para9_initTaskQueue() ================================================== */

typedef struct {
  int type; /* should be initialized by -1. */
  uint64_t bg; /* beginning index for the task. */
  uint64_t ed; /* ending index (exclusive) */
  int id;
} para9_Task;

/* This controls dependent and independent tasks assigned to threads.

   Queue status
    -1 : Initial.  Waiting for all the threads to be in while() loop.
     0 : All the threads are comming in the while() loop.
     1 : [OptimizerA] assinged and waiting for it to be finished.
 */
void para9_getTask(para9_TaskQueue * q, int tid, para9_Task * task){
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "{%d} getTask(); Old task=%d; Queue status=%d\n",
          tid, task->type, q->status);
  fflush(ytLib_ERR);
#endif

  int newTask = -1;

  /* ----------------------------------------------------------------------- */
  /* Progresses the status based on the task finished.                       */
  /* ----------------------------------------------------------------------- */
  if(task->type == -1){
    /* Wait for the final thread comming here */
	q->finishedThreads++;
	if(q->finishedThreads == q->threads){
	  q->status = 0; /* [initial status]  */
	}
  } else if(task->type == 1){  /* [OptimizerA] finished. */
    if(q->MPI_rank == 0){
      q->status = 20;
    } else {
      q->status = 23;
    }
  } else if(task->type == 10 || task->type == 11){
    /* [OptimizerF] or [OptimzerQ] finished. */
    q->finishedThreads++;
    if(q->finishedThreads == 2){
      q->status = 23; /* going to perform scatter. */
    }
  } else if(task->type == 12){
    /* [OptimizerB] finished. */
    q->status = 2;
  } else if(task->type == 2 || task->type == 3){
	/* [Send] or [Calc] finished */
	q->finishedThreads++;
	if(q->finishedThreads == q->threads + 1){
	  q->status = 5; /* [Calc] all finished. */
	}
  }

  /* ----------------------------------------------------------------------- */
  /* Assigns tasks based on status.                                          */
  /* ----------------------------------------------------------------------- */
  if(q->status == 0){
    if(q->MPI_rank < q->MPI_size){
      /* Initial status. Assign [OptimizerA]. */
      newTask = 1; 
      q->status = 1;
    } else {
      q->status = 999; /* finished */
    }

  } else if(q->status == 20){
    newTask = 10; /* OptimizerF. Only root */
    q->status = 21;
    q->assignedThreads = 1;
    q->finishedThreads = 0;

  } else if(q->status == 21){
    newTask = 11; /* OptimizerQ. only root */
    q->status = 22;
    q->assignedThreads = 2;
    
  } else if(q->status == 23){
    newTask = 12; /* OptimizerB */
    q->status = 24;
    
  } else if(q->status == 2){
	newTask = 2; /* [Send] */
	q->status = 3; /* [Send] assigned. */
	q->assignedThreads = 0;
	q->finishedThreads = 0;

  } else if(q->status == 3){
	newTask = 3; /* [Calc] */
	task->id = q->assignedThreads++;
	if(q->assignedThreads == q->threads){
	  q->status = 4; /* [Calc] all assigned. */
	}

  } else if(q->status == 5){
    /* [Calc] all finished. Check the finish condition */
    q->td->Fprevsrc = q->td->dstsrc[3];
    q->td->Qprevsrc = q->td->dstsrc[1];

    if((q->td->Qi >= ytIntArray_size(q->td->Qdst) &&
        q->td->dstsrc[0] == -1 && q->td->dstsrc[1] == -1 && q->td->Qprevsrc == -1) 
       && (q->td->Fi >= ytIntArray_size(q->td->Fdst) &&
        q->td->dstsrc[2] == -1 && q->td->dstsrc[3] == -1 && q->td->Fprevsrc == -1)){
#ifdef PARAOS_DEBUG
      fprintf(ytLib_ERR, "      Finish condition satisfied.\n");
      fflush(ytLib_ERR);
#endif
      /* Finish condition satisfied */
      q->status = 999; /* finished */
      ytTimer_clearStart(q->td->D_timer);

    } else {
      int ret;
      if(q->td->testNum != 0){
#ifdef PARAOS_DEBUG
        fprintf(ytLib_ERR, "{%d}   Calling MPI_Testall(); testNum=%d\n", tid, q->td->testNum);
        fflush(ytLib_ERR);
#endif
      /* MPI_Testall */
        MPI_Testall(q->td->testNum, q->td->testReqs, &ret, MPI_STATUSES_IGNORE);
      } else {
        /* No need to receive */
        ret = 1;
      }

      if(ret){
        if(q->td->testNum != 0){
          ytDoubleArray_add(q->td->C_timers + q->a - 1, ytTimer_get(q->td->C_timer));
        }
#ifdef PARAOS_DEBUG
		if(q->td->testNum != 0){
		  fprintf(ytLib_ERR, "{%d}   Communication finished.", tid);
		  if(q->td->dstsrc[2] != -1){
			int r = Util_readr_int32(((uint8_t *) q->db->F[q->Fb->send]) + q->td->max_F);
			fprintf(ytLib_ERR, "  F send rank=%d", r);
		  }
		  if(q->td->dstsrc[3] != -1){
			int r = Util_readr_int32(((uint8_t *) q->db->F[q->Fb->recv]) + q->td->max_F);
			fprintf(ytLib_ERR, "  F recv rank=%d", r);
		  }
		  if(q->td->dstsrc[0] != -1){
			int r = Util_readr_int32(q->db->Q[q->Qb->send] + q->td->max_Q);
			fprintf(ytLib_ERR, "  Q send rank=%d", r);
		  }
		  if(q->td->dstsrc[1] != -1){
			int r = Util_readr_int32(q->db->Q[q->Qb->recv] + q->td->max_Q);
			fprintf(ytLib_ERR, "  Q recv rank=%d", r);
		  }
		  fprintf(ytLib_ERR, "\n");
		  fflush(ytLib_ERR);
		}
#endif
        if(q->td->dstsrc[2] != -1 || q->td->dstsrc[3] != -1){
          int tmp = q->Fb->recv;
          q->Fb->recv = q->Fb->send;
          q->Fb->send = tmp;
        }
        if(q->td->dstsrc[0] != -1 || q->td->dstsrc[1] != -1){
          int tmp = q->Qb->recv;
          q->Qb->recv = q->Qb->send;
          q->Qb->send = tmp;
        }
        
        newTask = 2; /* [Send] */
        q->status = 3; /* [Send] assigned */
        q->assignedThreads = 0;
        q->finishedThreads = 0;
      }
    }
  }

  /* ----------------------------------------------------------------------- */
  /* Score calculation                                                       */
  /* ----------------------------------------------------------------------- */
  if(newTask == -1){
    if(q->a + 1 < q->p && q->scorePos != q->scoreTotal){
      /* Gives score calculation tasks if a + a < p */
      newTask = 0;
      uint64_t assign = (q->scoreTotal - q->scorePos) / q->g;
      if(assign == 0) assign = 1;
      
      task->bg = q->scorePos;
      q->scorePos += assign;
      
      if(q->scoreTotal < q->scorePos) q->scorePos = q->scoreTotal;
      task->ed = q->scorePos;
    } else if(q->status != 999){
      newTask = 999; /* Do nothing */
    }
  }
  
  task->type = newTask;
  
#ifdef PARAOS_DEBUG
  fprintf(ytLib_ERR, "{%d} New task=%d, id=%d, Queue status=%d\n", 
          tid, task->type, task->id, q->status);
  fflush(ytLib_ERR);
#endif
}/* para9_getTask() ======================================================== */

void para9_calcScores(const int tid, 
                      const uint64_t FNS, const uint64_t FN_bg, 
                      const uint64_t bg, const uint64_t ed,
                      const int p, const int a, void * sb,
                      para9_DB * db, int F_next){
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} calcScores() called; a=%d, FN_bg=%llu, bg=%llu, ed=%llu, F_next=%d.\n",
            tid, a, FN_bg, bg, ed, F_next);
    fflush(ytLib_ERR);
  }
#endif
  int A[PARAOS_MAX_NODES];
  for(uint64_t ii = bg; ii < ed; ii++){
    uint64_t i = ii + FN_bg;
    int v = i / FNS;
    uint64_t fs_id = i % FNS;
    ytCombo_combination2(p - 1, a, fs_id, A, db->R[v], db->table, p);
    float sc = Paraos_score(sb, v, A, a);
    para9_writeF(db->F[F_next] + ii, v, fs_id, a, sc);

#ifdef PARAOS_DUMPX
#pragma omp critical
    {
      fprintf(ytLib_ERR, "{%d} F[%d][%llu] = ", tid, F_next, ii);
      fprintf(ytLib_ERR, "F(%d,", v);
      ytCombo_print(ytLib_ERR, a, A);
      fprintf(ytLib_ERR, "):%llu <- s(%d,", i, v);
      ytCombo_print(ytLib_ERR, a, A);
      fprintf(ytLib_ERR, ") = %f\n", sc);
      fflush(ytLib_ERR);
    }
#endif
  }/* for FN_bg < FN_ed; */
}/* para9_calcScores() ===================================================== */

void para9_taskOptimizerA(const int MPI_rank, const int MPI_size,
                          const int Fsend, const int Qsend,
                          para9_ThreadData * td, para9_DB * db,
                          int threads, int a){
  td->Fprevsrc = -1;
  td->Qprevsrc = -1;
  ytTimer_clearStart(db->M_timer); /* Start of M_timer */
  para9_calcSrc(MPI_rank, MPI_size, threads,
				db->Fflag, db->Qflag, db->QFflag,
				td->Fdst, td->Qdst,
				&(td->Fprevsrc), &(td->Qprevsrc));
  
  ytTimer_sector(db->M_timer); /* 0 */
  
  para9_optimizeGather(MPI_rank, MPI_size, td->Fdst, td->Qdst, db, a);
}/* para9_taskOptimizerA() ================================================ */

void para9_taskOptimizerB(const int MPI_rank, const int MPI_size,
                          const int Fsend, const int Qsend,
                          para9_ThreadData * td, para9_DB * db,
                          int threads, int a){

  para9_optimizer(MPI_rank, MPI_size, td->Fdst, td->Qdst, db, a);

  /* Writes the source ranks at the buffer footer */
  Util_write_int32(((uint8_t *) db->F[Fsend]) + td->max_F, MPI_rank);
  Util_write_int32(db->Q[Qsend] + td->max_Q, MPI_rank);

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
	int tid = omp_get_thread_num();
	int r = Util_readr_int32(db->Q[Qsend] + td->max_Q);
	fprintf(ytLib_ERR, "{%d} Wrote the Q source rank (%d) at %llu, send=%d\n",
			tid, r, td->max_Q, Qsend);
	fflush(ytLib_ERR);
  }
#endif
  
  td->Fi = 0;
  td->Qi = 0;
}/* para9_taskOptimizerB() ================================================= */

void para9_taskSend(const int p, const int a, const int MPI_rank, 
                    const int Fsend, const int Frecv,
                    const int Qsend, const int Qrecv,
                    para9_ThreadData * td, para9_DB * db, const int tid){
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
	fprintf(ytLib_ERR, "{%d} [Send] Qprevsrc=%d, Fprevsrc=%d(%d)\n", 
			tid, td->Qprevsrc, 
			td->Fprevsrc == -1 ? -1 : td->Fprevsrc & 0xffffff, td->Fprevsrc >> 24);
	fflush(ytLib_ERR);
  }
#endif
  
  /* F --------------------------------------------------------------------- */
  if(td->Fi < ytIntArray_size(td->Fdst)){
	td->dstsrc[2] = ytIntArray_get(td->Fdst, td->Fi);
	td->dstsrc[3] = ytIntArray_get(td->Fdst, td->Fi + 1);
	td->Fi += 2;
	if((td->dstsrc[3] & 0xffffff) == td->dstsrc[2] && td->dstsrc[2] == MPI_rank){
	  td->dstsrc[2] = -1; td->dstsrc[3] = -1;
	}
	
	/* F send */
	if(td->dstsrc[2] != -1){
#ifdef PARAOS_DEBUG
#pragma omp critical
	  {
		fprintf(ytLib_ERR, "{%d} [F] MPI_Isend(); send=%d, dst=%d, tag=%d, size=%d\n", 
				tid, Fsend, td->dstsrc[2], a + p, td->FsendSize);
		fflush(ytLib_ERR);
	  }
#endif

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
	fprintf(ytLib_ERR, "{%d} dstsrc={%d,%d,%d,%d}\n",
			tid,td->dstsrc[0], td->dstsrc[1], td->dstsrc[2], td->dstsrc[3]);
	fflush(ytLib_ERR);
  }
#endif
	  MPI_Isend(db->F[Fsend], td->FsendSize, MPI_UINT32_T,
				//td->dstsrc[2], a + p, MPI_COMM_WORLD, &(td->reqs[2]));
                td->dstsrc[2], a + p, db->comm[a - 1], &(td->reqs[2]));
	}
	/* F receive */
	if(td->dstsrc[3] != -1){
#ifdef PARAOS_DEBUG
#pragma omp critical
	  {
		fprintf(ytLib_ERR, "{%d} [F] MPI_Irecv(): recv=%d, src=%d, tag=%d, size=%d\n", 
				tid, Frecv, td->dstsrc[3] & 0xffffff, a + p, td->FsendSize);
		fflush(ytLib_ERR);
	  }
#endif
	  MPI_Irecv(db->F[Frecv], td->FsendSize, MPI_UINT32_T, td->dstsrc[3] & 0xffffff, a + p,
				//MPI_COMM_WORLD, &(td->reqs[3]));
                db->comm[a - 1], &(td->reqs[3]));
	}
  } else {
	td->dstsrc[2] = -1;
	td->dstsrc[3] = -1;
  }

  /* Q --------------------------------------------------------------------- */
  if(td->Qi < ytIntArray_size(td->Qdst)){
	td->dstsrc[0] = ytIntArray_get(td->Qdst, td->Qi);
	td->dstsrc[1] = ytIntArray_get(td->Qdst, td->Qi + 1);
	td->Qi += 2;
	/* Skip condition */
	if((td->dstsrc[1] & 0xffffff) == td->dstsrc[0] && td->dstsrc[0] == MPI_rank){
	  td->dstsrc[0] = -1; td->dstsrc[1] = -1;
	}
	/* Q send */
	if(td->dstsrc[0] != -1){
#ifdef PARAOS_DEBUG
#pragma omp critical
	  {
		fprintf(ytLib_ERR, "{%d} [Q] MPI_Isend(); send=%d, dst=%d, tag=%d, size=%d (%d bytes)\n", 
				tid, Qsend, td->dstsrc[0], a, td->QsendSize, td->QsendSize * 4);
		fflush(ytLib_ERR);
	  }
#endif

#ifdef PARAOS_DEBUG
#pragma omp critical
  {
	fprintf(ytLib_ERR, "{%d} dstsrc={%d,%d,%d,%d}\n",
			tid,td->dstsrc[0], td->dstsrc[1], td->dstsrc[2], td->dstsrc[3]);
	fflush(ytLib_ERR);
  }
#endif

	  MPI_Isend(db->Q[Qsend], td->QsendSize, MPI_UINT32_T, td->dstsrc[0], a, 
                //MPI_COMM_WORLD, &(td->reqs[0]));
                db->comm[a - 1], &(td->reqs[0]));
	}
	/* Q receive */
	if(td->dstsrc[1] != -1){
#ifdef PARAOS_DEBUG
#pragma omp critical
	  {
		fprintf(ytLib_ERR, "{%d} [Q] MPI_Irecv(): recv=%d, src=%d, tag=%d, size=%d (%d bytes)\n",
				tid, Qrecv, td->dstsrc[1] & 0xffffff, a, td->QsendSize, td->QsendSize * 4);
		fflush(ytLib_ERR);
	  }
#endif
	  MPI_Irecv(db->Q[Qrecv], td->QsendSize, MPI_UINT32_T, td->dstsrc[1] & 0xffffff, a,
				//MPI_COMM_WORLD, &(td->reqs[1]));
                db->comm[a - 1], &(td->reqs[1]));
	}
#ifdef PARAOS_DEBUG
	else {
#pragma omp critical
	  {
		fprintf(ytLib_ERR, "{%d} [Q] no receive required\n", tid);
		fflush(ytLib_ERR);
	  }
	}
#endif
  } else {
	td->dstsrc[0] = -1;
	td->dstsrc[1] = -1;
  }
  
  td->testNum = 0;
  for(int i = 0; i < 4; i++){
	if(td->dstsrc[i] != -1){
	  td->testIDs[td->testNum] = i;
	  td->testReqs[td->testNum++] = td->reqs[i];
	}
  }
#ifdef PARAOS_DEBUG
#pragma omp critical
  {
    fprintf(ytLib_ERR, "{%d} task type=2 [Send] finished.\n", tid); fflush(ytLib_ERR);
  }
#endif
}/* para9_taskSend() ======================================================= */

void para9_dump_F(int p, int a, int buff, int MPI_rank, int MPI_size,
                  uint64_t minimum, const para9_DB * db){
  uint64_t total_FS = ytCombo_chooseTable(p - 1, a, db->table, p);
  para9_Index F;
  para9_calcIndex(&F, total_FS * p, MPI_size, minimum);

  fprintf(ytLib_ERR, "F.total=%llu, F.quot=%llu, F.rem=%llu\n", F.total, F.quot, F.rem);
  fprintf(ytLib_ERR, "rank=%d, size=%d, F.size=%llu\n", MPI_rank, MPI_size, F.size);
  fflush(ytLib_ERR);

  uint64_t bg, ed, elm;
  para9_splitJobs(&F, MPI_rank, &bg, &ed, &elm);

  fprintf(ytLib_ERR, "bg=%llu, ed=%llu, elm=%llu\n", bg, ed, elm);
  fflush(ytLib_ERR);

  int A[a];
  for(uint64_t ii = 0; ii < elm; ii++){
    int v = (ii + bg) / total_FS;
    uint64_t id = (ii + bg) % total_FS;
    float sc;
    uint64_t B;
    uint8_t len;
    uint8_t w;
    para9_readF(db->F[buff] + ii, &w, &B, &len, &sc);

    //fprintf(ytLib_ERR, "B=%llu len=%d ", B, len);

    fprintf(ytLib_ERR, "F[%d][%llu]=F(%d,", buff, ii, v);
    fflush(ytLib_ERR);
    ytCombo_combination2(p - 1, a, id, A, db->R[v], db->table, p);
    ytCombo_print(ytLib_ERR, a, A);
    fflush(ytLib_ERR);
    fprintf(ytLib_ERR, ":%llu)=s(%d,", id, w);
    fflush(ytLib_ERR);
    ytCombo_combination2(p - 1, len, B, A, db->R[v], db->table, p);
    //fprintf(ytLib_ERR, "D"); fflush(ytLib_ERR);
    ytCombo_print(ytLib_ERR, len, A);
    fprintf(ytLib_ERR, ")=%f\n", sc);
  }
  fflush(ytLib_ERR);
}/* para9_dump_F() ========================================================= */

void para9_dump_Q(int p, int a, int buff, int MPI_rank, int MPI_size,
                  uint64_t minimum, const para9_DB * db){
  para9_Index Q;
  para9_calcIndex(&Q, ytCombo_chooseTable(p, a, db->table, p),
                  MPI_size, minimum);

  uint8_t * pos = db->Q[buff];
  uint64_t bg, ed, elm;
  para9_splitJobs(&Q, MPI_rank, &bg, &ed, &elm);

  fprintf(ytLib_ERR, "Q.total=%llu, Q.size=%llu\n", Q.total, Q.size);
  fprintf(ytLib_ERR, "bg=%llu, ed=%llu, elm=%llu\n", bg, ed, elm);
  fflush(ytLib_ERR);

  int A[a];
  for(uint64_t ii = 0; ii < elm; ii++){
    ytCombo_combination(p, a, ii + bg, A, db->table, p);
    fprintf(ytLib_ERR, "Q[%d][%llu]=Q(", buff, ii); fflush(ytLib_ERR);
    ytCombo_print(ytLib_ERR, a, A); fflush(ytLib_ERR);
    fprintf(ytLib_ERR, ")="); fflush(ytLib_ERR);

    float sc = Util_readr_float(pos);
    pos += sizeof(float);

    for(int j = 0; j < a; j++){
      uint32_t Bl;
      memcpy(&Bl, pos, sizeof(uint32_t));
      pos += sizeof(uint32_t);
      //fprintf(ytLib_ERR, "Bl=%u,", Bl); fflush(ytLib_ERR);
      uint16_t Bu;
      memcpy(&Bu, pos, sizeof(uint16_t));
      //fprintf(ytLib_ERR, "Bu=%u,", Bu); fflush(ytLib_ERR);
      pos += sizeof(uint16_t);
      uint64_t B = (uint64_t) Bl | ((uint64_t) Bu) << 32;
      //fprintf(ytLib_ERR, "B=%llu,", B); fflush(ytLib_ERR);
      uint8_t v;
      memcpy(&v, pos, sizeof(uint8_t));
      //fprintf(ytLib_ERR, "v=%u,", v); fflush(ytLib_ERR);
      pos += sizeof(int8_t);
      uint8_t len;
      memcpy(&len, pos, sizeof(uint8_t));
      pos += sizeof(uint8_t);
      ytCombo_combination2(p - 1, len, B, A, db->R[v], db->table, p);
      if(j != 0) fprintf(ytLib_ERR, "+");
      fprintf(ytLib_ERR, "s(%d,", v); fflush(ytLib_ERR);
      ytCombo_print(ytLib_ERR, len, A);
      fprintf(ytLib_ERR, ")"); fflush(ytLib_ERR);
    }
    fprintf(ytLib_ERR, " = %f\n", sc);
  }
  fflush(ytLib_ERR);
}/* para9_dump_Q() ========================================================= */

ytPCGraph * para9_graph(int p, const uint8_t * Q, const para9_DB * db){
  ytPCGraph * g = ytPCGraph_new(p);
  float sc = Util_readr_float(Q);

  Q += sizeof(float);

  fprintf(ytLib_ERR, "Optimal score = %f\n", sc);
  fprintf(ytLib_ERR, "X-OPTIMALSCORE\t%f\n", sc);
  fprintf(ytLib_ERR, "Optimal pi = (");
  fflush(ytLib_ERR);
  const uint8_t * P = Q;
  for(int i = 0; i < p; i++){
    P += sizeof(int32_t) + sizeof(int16_t);
    int v = Util_readr_int8(P);
    P += sizeof(int8_t) + sizeof(int8_t);
    if(i != 0) fprintf(ytLib_ERR, ",");
    fprintf(ytLib_ERR, "%d", v);
  }
  fprintf(ytLib_ERR, ")\n");

  int A[p];
  for(int i = 0; i < p; i++){
    uint32_t Bl;
    memcpy(&Bl, Q, sizeof(uint32_t));
    Q += sizeof(uint32_t);
    uint16_t Bu;
    memcpy(&Bu, Q, sizeof(uint16_t));
    Q += sizeof(uint16_t);
    uint64_t B;
    B = (uint64_t) Bl | ((uint64_t) Bu) << 32;
    int8_t v;
    memcpy(&v, Q, sizeof(int8_t));
    Q += sizeof(int8_t);
    int8_t len;
    memcpy(&len, Q, sizeof(int8_t));
    Q += sizeof(int8_t);

    ytCombo_combination2(p - 1, len, B, A, db->R[v], db->table, p);
    for(int j = 0; j < len; j++){
      ytPCGraph_addEdge(g, A[j], v);
    }
  }
  return g;
}/* para9_graph() ========================================================== */

void parallel9_mem_test(Args * args, int p, int procs){
  para9_DB * db = malloc(sizeof(para9_DB));
  size_t base = sizeof(para9_DB);
  int threads = omp_get_max_threads();
  if(args->dry_threads != -1) threads = args->dry_threads;
  uint64_t minElms = (uint64_t) args->minElms;
  size_t total = para9_alloc(p, procs, 0, db, threads, 1, minElms) + base;

  fprintf(ytLib_ERR, "\n=====================\n");
  fprintf(ytLib_ERR, "Total Memory:\n");
  ytFile_printComma_size_t(ytLib_ERR, total);
  fprintf(ytLib_ERR, " [B]\n");
  ytFile_printComma_size_t(ytLib_ERR, total >> 10);
  fprintf(ytLib_ERR, " [KiB]\n");
  ytFile_printComma_size_t(ytLib_ERR, total >> 20);
  fprintf(ytLib_ERR, " [MiB]\n");
  ytFile_printComma_size_t(ytLib_ERR, total >> 30);
  fprintf(ytLib_ERR, " [GiB]\n");

}/* parallel9_mem_test() =================================================== */

/*****************************************************************************/
/* MAIN ROUTINE                                                              */
/*****************************************************************************/

ytPCGraph * parallel9(Args * args, int p, void ** sb){
  int threads = omp_get_max_threads();
  /* Minimum elements */
  uint64_t minElms = (uint64_t) args->minElms;
  
  fprintf(ytLib_ERR, "Parallel Optimal DAG Search Algorithm Version 9.\n");
  fprintf(ytLib_ERR, "  p=%d\n", p);
  fprintf(ytLib_ERR, "  MPI size=%d\n", args->MPI_size);
  fprintf(ytLib_ERR, "  OpenMP Threads=%d\n", threads);
  fprintf(ytLib_ERR, "  minElms=");
  if(minElms == UINT64_MAX){
    fprintf(ytLib_ERR, "UNLIMITED\n");
  } else {
    fprintf(ytLib_ERR, "%llu\n", minElms);
  }
  fflush(ytLib_ERR);

  para9_DB * db = malloc(sizeof(para9_DB));
  size_t total_mem = para9_alloc(p, args->MPI_size, args->MPI_rank, db, threads, 0, minElms);

  fprintf(ytLib_ERR, "==========================\n");
  fprintf(ytLib_ERR, "Total Memory:\n");
  ytFile_printComma_size_t(ytLib_ERR, total_mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  ytFile_printComma_size_t(ytLib_ERR, total_mem >> 30);
  fprintf(ytLib_ERR, " [GB]\n");
  fprintf(ytLib_ERR, "==========================\n");

  ytDoubleArray * L_timers = ytDoubleArray_new();
  ytTimer * L_timer = ytTimer_new(); /* lap timer */
  ytDoubleArray * T_timers = ytDoubleArray_arrayNew(p); /* thread timer */
  ytTimer * T_timer[threads];
  for(int tid = 0; tid < threads; tid++){
    T_timer[tid] = ytTimer_new();
  }
  ytDoubleArray * P_timers = ytDoubleArray_new();
  ytTimer * P_timer = ytTimer_new(); /* path optimization timer */
  ytDoubleArray * R_timers = ytDoubleArray_new();
  ytTimer * R_timer = ytTimer_new();
  ytDoubleArray * G_timers = ytDoubleArray_new();
  ytTimer * G_timer = ytTimer_new();
  db->M_timer = ytTimer_new();
  ytTimer_enableSector(db->M_timer);

  para9_Index F, FM, FN, QM, Q, QN;
  uint64_t FS, FMS, FNS;

  /* ----------------------------------------------------------------------- */
  /* For |A| = 0                                                             */
  /* ----------------------------------------------------------------------- */
  fprintf(ytLib_ERR, 
          "============================================================\n");
  fprintf(ytLib_ERR, "|A|=0\n");
  fflush(ytLib_ERR);

  FS = 1;
  FNS = ytCombo_chooseTable(p - 1, 1, db->table, p);
  para9_calcIndex(&F, p, args->MPI_size, minElms);
  para9_calcIndex(&FN, FNS * p, args->MPI_size, minElms);
  para9_calcIndex(&Q, 0, args->MPI_size, minElms);
  para9_calcIndex(&QN, ytCombo_chooseTable(p, 1, db->table, p), args->MPI_size, minElms);

  uint64_t F_bg, F_ed, F_elm;
  para9_splitJobs(&F, args->MPI_rank, &F_bg, &F_ed, &F_elm);

  uint64_t FN_bg, FN_ed, FN_elm;
  para9_splitJobs(&FN, args->MPI_rank, &FN_bg, &FN_ed, &FN_elm);

  uint64_t Q_bg = 0, Q_ed, Q_elm;

  db->Qsize = para9_getQSize(0);

#ifdef PARAOS_DUMP
  fprintf(ytLib_ERR, "F.total=%llu, F_bg=%llu, F_elm=%llu\n", F.total, F_bg, F_elm);
  fflush(ytLib_ERR);
#endif
  
  db->B[1][0][0] = 0;
  para9_TaskQueue * queue = malloc(sizeof(para9_TaskQueue));

  /* Initial buffers : {send, recv, calc, next} */
  para9_BuffIndex Fb = {0, 1, 2,  3};
  para9_BuffIndex Qb = {0, 1, 2, -1};

  para9_ThreadData * td = malloc(sizeof(para9_ThreadData));
  td->Fdst = ytIntArray_new();
  td->Qdst = ytIntArray_new();
  td->C_timer = ytTimer_new();
  td->C_timers = ytDoubleArray_arrayNew(p);
  td->D_timer = ytTimer_new();
  td->D_timers = ytDoubleArray_new();

  /*-------------------------------------------------*/
  /* Calculates s(v,0) = F(v,0) & s(v,A) where |A|=1 */
  /*-------------------------------------------------*/
#pragma omp parallel
  {
    int tid = omp_get_thread_num();
    uint64_t quot = F_elm / threads;
    uint64_t rem =  F_elm % threads;
    uint64_t bg, ed, elm;
    Util_splitJobs2(F_elm, tid, quot, rem, &bg, &ed, &elm);

    para9_calcScores(tid, FS, F_bg, bg, ed,
                     p, 0, sb[tid], db, Fb.send);

    quot = FN_elm / threads;
    rem =  FN_elm % threads;
    Util_splitJobs2(FN_elm, tid, quot, rem, &bg, &ed, &elm);

    para9_calcScores(tid, FNS, FN_bg, bg, ed,
                     p, 1, sb[tid], db, Fb.calc);
  }/* omp parallel */

#ifdef PARAOS_DUMP
  fprintf(ytLib_ERR, "Calced local F()'s:\n");
  para9_dump_F(p, 0, Fb.send, args->MPI_rank, args->MPI_size, minElms, db);
#endif


  /***************************************************************************/
  /* MAIN LOOP                                                               */
  /***************************************************************************/
  /*
    At loop 'a', calculates F(v,A) where |A| = a, and Q(M(A)) where |A| = a,
    both from F(v,A') where |A'| = a - 1.
  */

  for(int a = 1; a <= p; a++){
    ytTimer_clearStart(L_timer);
    ytTimer_clearStart(R_timer);
    int slotB = (a + 1) % 2;

    fprintf(ytLib_ERR, 
            "============================================================\n");
    fprintf(ytLib_ERR, "|A|=%d\n", a);
#ifdef PARAOS_DEBUG
    int slotA = a % 2;
    fprintf(ytLib_ERR, "  slotA=%d, slotB=%d\n", slotA, slotB);
#endif
#ifdef PARAOS_GTIME
    fprintf(ytLib_ERR, "X-GTIME:\t%d\t%d\t%d\t%.6f\n", 
            args->MPI_rank, a, 0, ytTimer_getTime());
#endif
    fflush(ytLib_ERR);

    /* Updates indexes ----------------------------------------------------- */
    para9_Index FT, QT;
    FMS = FS;
    FS = FNS;
    FM = F;
    F = FN;
    F_bg = FN_bg;
    F_ed = FN_ed;
    F_elm = FN_elm;
    QM = Q;
    Q = QN;
    if(a < p){
      FNS = ytCombo_chooseTable(p - 1, a + 1, db->table, p);
      para9_calcIndex(&FN, FNS * p, args->MPI_size, minElms);
      para9_splitJobs(&FN, args->MPI_rank, &FN_bg, &FN_ed, &FN_elm);
	  para9_calcIndex(&FT, F_elm, threads, 0);
      para9_calcIndex(&QN, ytCombo_chooseTable(p, a + 1, db->table, p), args->MPI_size, minElms);
    }

    para9_splitJobs(&Q, args->MPI_rank, &Q_bg, &Q_ed, &Q_elm);
    
    para9_calcIndex(&QT, Q_elm, threads, 0);

    td->max_F = FM.quot;
    if(FM.rem != 0) td->max_F += 1;
    td->max_F *= para9_getFSize();

    td->max_Q = QM.quot;
    if(QM.rem != 0) td->max_Q += 1;
    td->max_Q *= para9_getQSize(a - 1);

    td->FsendSize = td->max_F / 4 + 1; /* in 4 bytes */
    td->QsendSize = td->max_Q / 4 + 1; /* in 4 bytes */

    db->Qtmp_elm_size = max(para9_getQSize(a), para9_getFSize());
    db->QMsize = db->Qsize;
    db->Qsize = para9_getQSize(a);


#ifdef PARAOS_DEBUG
    fprintf(ytLib_ERR, "Q.total=%llu, Q.size=%llu, Q_bg=%llu, Q_ed=%llu, Q_elm=%llu\n", 
            Q.total, Q.size, Q_bg, Q_ed, Q_elm);
    fprintf(ytLib_ERR, "QM.total=%llu\n", QM.total);
    fprintf(ytLib_ERR, "QT.total=%llu, QT.quot=%llu, QT.rem=%llu, QT.size=%llu\n",
            QT.total, QT.quot, QT.rem, QT.size);
    if(a < p){
      fprintf(ytLib_ERR, "FN.total=%llu, FN.quot=%llu, FN.rem=%llu, FN.size=%llu\n",
              FN.total, FN.quot, FN.rem, FN.size);
      fprintf(ytLib_ERR, "F.total=%llu, F.size=%llu\n", F.total, F.size);
      fprintf(ytLib_ERR, "FN_bg=%llu, FN_ed=%llu, FN_elm=%llu\n", FN_bg, FN_ed, FN_elm);
      fprintf(ytLib_ERR, "F_bg=%llu, F_ed=%llu, F_elm=%llu\n", F_bg, F_ed, F_elm);
    }
    fflush(ytLib_ERR);
#endif

    /* Generates sub combinations ------------------------------------------ */
    if(a < p){
#ifdef PARAOS_DUMP
      fprintf(ytLib_ERR, "Generating sub-combinations\n");
#endif
      for(int i = 0; i <= a; i++){
        ytCombo_combination(a + 1, a, i, db->B[slotB][i], db->table, p);
#ifdef PARAOS_DUMP
        fprintf(ytLib_ERR, "B[%d][%d]=", slotB, i);
        ytCombo_print(ytLib_ERR, a, db->B[slotB][i]);
        fprintf(ytLib_ERR, "\n");
        fflush(ytLib_ERR);
#endif
      }
    }/* if(a < p) */

    para9_initTaskQueue(queue, p, a, threads, FN_elm, args->g, td, &Fb, &Qb,
                        args->MPI_rank, db->commSize[a - 1]);
#ifdef PARAOS_DEBUG
	queue->db = db;
#endif
    
	/*====================================================================== */
    /* PARALLEL REGION ===================================================== */
#pragma omp parallel
    {
      int tid = omp_get_thread_num();
      ytTimer_clearStart(T_timer[tid]);
      para9_Task task = {-1, 0, 0};
      uint64_t FT_bg, FT_ed, FT_elm;
      uint64_t QT_bg, QT_ed, QT_elm;
      Util_splitJobs2(F_elm, tid, FT.quot, FT.rem, &FT_bg, &FT_ed, &FT_elm);
      Util_splitJobs2(Q_elm, tid, QT.quot, QT.rem, &QT_bg, &QT_ed, &QT_elm);

#ifdef PARAOS_DEBUG
#pragma omp critical
      {
        fprintf(ytLib_ERR, "{%d} QT_bg=%llu, QT_ed=%llu, QT_elm=%llu\n", tid, 
                QT_bg, QT_ed, QT_elm); fflush(ytLib_ERR);
      }
#endif

      /* Clear Fflag */
      for(int i = 0; i < args->MPI_size; i++) db->Fflag[tid][i] = 0;
      /*   Note: Qflag and QFflag are cleared in the prepare_Q(). */

      /* prepare F */
      if(a < p){
        para9_prepare_F(p, a, db->commSize[a - 1],
                        &FM, FMS, FS, F_bg, FT_bg, FT_ed, FT_elm,
                        db, tid, threads);
      }/* if(a < p) */

      para9_prepare_Q(p, a, db->commSize[a - 1], &QM, &FM, FMS,
                      Q_bg, QT_bg, QT_ed, QT_elm, db, tid, threads);

      /* Initialization of Qtmp[] and Q[]. */
      for(uint64_t i = QT_bg * a; i < QT_ed * a; i++){
        size_t addr = i * db->Qtmp_elm_size;
        *((float *) (db->Qtmp + addr)) = 0;
      }
      for(uint64_t i = QT_bg; i < QT_ed; i++){
        size_t addr = i * db->Qsize;
        *((float *) (db->Q[Qb.calc] + addr)) = FLT_MAX;
      }

      while(1){
#pragma omp critical
        {
          ytTimer_start(G_timer);
          para9_getTask(queue, tid, &task);
          ytTimer_stop(G_timer);
        }

        if(task.type == -1){
          /* Finished ------------------------------------------------------ */
#ifdef PARAOS_DEBUG
#pragma omp critical
          {
            fprintf(ytLib_ERR, "{%d} Task finished.\n", tid);
            fflush(ytLib_ERR);
          }
#endif
          break;
          
        } else if(task.type == 0){
          /* [Score calculation] ------------------------------------------- */
          para9_calcScores(tid, FNS, FN_bg, task.bg, task.ed, p, a + 1, 
                           sb[tid], db, Fb.next);

        } else if(task.type == 1){
          /* [OptimizerA] -------------------------------------------------- */
#ifdef PARAOS_GTIME
          /* GTIME ID=1 */
          fprintf(ytLib_ERR, "X-GTIME:\t%d\t%d\t%d\t%.6f\n", 
                  args->MPI_rank, a, 1, ytTimer_getTime());
          fflush(ytLib_ERR);
#endif
          ytDoubleArray_add(R_timers, ytTimer_get(R_timer));
          ytTimer_clearStart(P_timer);
          para9_taskOptimizerA(args->MPI_rank, db->commSize[a - 1],
                               Fb.send, Qb.send, td, db, threads, a);
          ytTimer_sector(db->M_timer); /* 4 */
        } else if(task.type == 10){
          /* [OptimizerF] -------------------------------------------------- */
          para9_optimizerF(args->MPI_rank, db->commSize[a - 1],
                           td->Fdst, db, a);

        } else if(task.type == 11){
          /* [OptimizerQ] -------------------------------------------------- */
          para9_optimizerQ(args->MPI_rank, db->commSize[a - 1],
                           td->Qdst, db, a);

        } else if(task.type == 12){
          /* [OptimizerB] -------------------------------------------------- */
          ytTimer_sector(db->M_timer); /* 5 */
          para9_taskOptimizerB(args->MPI_rank, db->commSize[a - 1],
                               Fb.send, Qb.send, td, db, threads, a);
          ytDoubleArray_add(P_timers, ytTimer_get(P_timer));
          fprintf(ytLib_ERR, "Sector P_timers (Optimizer): %f\n", 
                  ytDoubleArray_get(P_timers, a - 1));
          fflush(ytLib_ERR);

#ifdef PARAOS_GTIME
          /* GTIME ID=3 */
          fprintf(ytLib_ERR, "X-GTIME:\t%d\t%d\t%d\t%.6f\n", 
                  args->MPI_rank, a, 3, ytTimer_getTime());
          fflush(ytLib_ERR);
#endif
        } else if(task.type == 2){
          /* [Send] -------------------------------------------------------- */
          ytTimer_clearStart(td->C_timer);
          para9_taskSend(p, a, db->commSize[a - 1],
                         Fb.send, Fb.recv, Qb.send, Qb.recv, td, db, tid);

        } else if(task.type == 3){
		  /* [Calc] -------------------------------------------------------- */

          if(td->Fprevsrc != -1){
            int orig = Util_readr_int32(((uint8_t *) db->F[Fb.send]) + td->max_F);
            uint64_t FMR_bg, FMR_ed, FMR_elm;
            para9_splitJobs(&FM, orig, &FMR_bg, &FMR_ed, &FMR_elm);
            
            switch(td->Fprevsrc >> 24){
            case 1: /* Updates only F */
              para9_calc_F(orig, Fb.send, Fb.calc, F_bg, FMR_bg, 
#ifdef PARAOS_DUMP
                           p, a, FS, FMS, 
#endif
                           db, task.id);
              break;

            case 2: /* updates only Q */
              para9_calc_QF(p, a, orig, Fb.send, Qb.calc,
                            Q_bg, FMR_bg, FMS, db, task.id);
              break;

            case 3: /* Both */
              para9_calc_F(orig, Fb.send, Fb.calc, F_bg, FMR_bg, 
#ifdef PARAOS_DUMP
                           p, a, FS, FMS, 
#endif
                           db, task.id);
              para9_calc_QF(p, a, orig, Fb.send, Qb.calc,
                            Q_bg, FMR_bg, FMS, db, task.id);
              break;
            }
          }
          if(!(td->Qprevsrc & 0x4000000)){
            int orig = Util_readr_int32(db->Q[Qb.send] + td->max_Q);
            uint64_t QMR_bg, QMR_ed, QMR_elm;
            para9_splitJobs(&QM, orig, &QMR_bg, &QMR_ed, &QMR_elm);
#ifdef PARAOS_DEBUG
#pragma omp critical
			{
			  fprintf(ytLib_ERR, "{%d} Calling calc_Q(); orig=%d at %llu\n",
					  tid, orig, td->max_Q);
			  fprintf(ytLib_ERR, "{%d}   QM.total=%llu\n",
					  tid, QM.total);
			  fflush(ytLib_ERR);
			}
#endif
            para9_calc_Q(orig, Qb.send, Qb.calc, Q_bg, QMR_bg, a,
#ifdef PARAOS_DUMP
                         p,
#endif
                         db, task.id);
          }
        }
      }/* while(1) */
      ytTimer_stop(T_timer[tid]);
#pragma omp critical
      {
        ytDoubleArray_set(T_timers + a - 1, tid, ytTimer_get(T_timer[tid]));
      }
    }/* omp parallel region */

    //fprintf(ytLib_ERR, "Finished for a=%d\n", a);
#ifdef PARAOS_DUMP
    fprintf(ytLib_ERR, "Calced local F()'s:\n");
    para9_dump_F(p, a, Fb.calc, args->MPI_rank, args->MPI_size, db);
    fprintf(ytLib_ERR, "Calced local Q()'s:\n");
    para9_dump_Q(p, a, Qb.calc, args->MPI_rank, args->MPI_size, db);
    fflush(ytLib_ERR);
#endif

    /* Swaps 'calc' and 'send' buffers */
    int tmp = Qb.send;
    Qb.send = Qb.calc;
    Qb.calc = tmp;

    tmp = Fb.send;
    Fb.send = Fb.calc;
    Fb.calc = Fb.next;
    Fb.next = tmp;

    ytTimer_stop(L_timer);
    ytDoubleArray_add(L_timers, ytTimer_get(L_timer));
    ytDoubleArray_add(td->D_timers, ytTimer_get(td->D_timer));
    ytDoubleArray_add(G_timers, ytTimer_get(G_timer));
    ytTimer_clear(G_timer);

    fprintf(ytLib_ERR, "Sector L_timers (Lap time) : %f\n",
            ytDoubleArray_get(L_timers, a - 1));
    fflush(ytLib_ERR);
  }/* for 1 <= a <= p */
  /***************************************************************************/
  /* End of main loop                                                        */
  /***************************************************************************/

  double sum = 0.0;
  fprintf(ytLib_ERR, "L_timers:\n");
  for(int i = 0; i < ytDoubleArray_size(L_timers); i++){
    sum += ytDoubleArray_get(L_timers, i);
    fprintf(ytLib_ERR, "%d: %f\n", i + 1, ytDoubleArray_get(L_timers, i));
  }
  fprintf(ytLib_ERR, "TOTAL: %f\n", sum);

  fprintf(ytLib_ERR, "T_timers:\n");
  para9_printTimers(p, T_timers);

  sum = 0.0;
  fprintf(ytLib_ERR, "P_timers (Optimizer):\n");
  for(int i = 0; i < ytDoubleArray_size(P_timers); i++){
    sum += ytDoubleArray_get(P_timers, i);
    fprintf(ytLib_ERR, "%d: %f\n", i + 1, ytDoubleArray_get(P_timers, i));
  }
  fprintf(ytLib_ERR, "TOTAL: %f\n", sum);

  fprintf(ytLib_ERR, "C_timers:\n");
  para9_printTimers(p, td->C_timers);

  sum = 0.0;
  fprintf(ytLib_ERR, "D_timers:\n");
  for(int i = 0; i < ytDoubleArray_size(td->D_timers); i++){
    sum += ytDoubleArray_get(td->D_timers, i);
    fprintf(ytLib_ERR, "%d: %f\n", i + 1, ytDoubleArray_get(td->D_timers, i));
  }
  fprintf(ytLib_ERR, "TOTAL: %f\n", sum);

  sum = 0.0;
  fprintf(ytLib_ERR, "R_timers:\n");
  for(int i = 0; i < ytDoubleArray_size(R_timers); i++){
    sum += ytDoubleArray_get(R_timers, i);
    fprintf(ytLib_ERR, "%d: %f\n", i + 1, ytDoubleArray_get(R_timers, i));
  }
  fprintf(ytLib_ERR, "TOTAL: %f\n", sum);

  sum = 0.0;
  fprintf(ytLib_ERR, "G_timers:\n");
  for(int i = 0; i < ytDoubleArray_size(G_timers); i++){
    sum += ytDoubleArray_get(G_timers, i);
    fprintf(ytLib_ERR, "%d: %f\n", i + 1, ytDoubleArray_get(G_timers, i));
  }
  fprintf(ytLib_ERR, "TOTAL: %f\n", sum);

  fprintf(ytLib_ERR, "M_timers:\n");
  //para9_printTimers(p, db->M_timers);
  ytTimer_print(db->M_timer, ytLib_ERR);

  size_t mem = ytIntArray_arrayMemSize(db->dstTableF, args->MPI_size);
  fprintf(ytLib_ERR, "dstTableF size = ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  
  mem = ytIntArray_arrayMemSize(db->dstTableQ, args->MPI_size);
  fprintf(ytLib_ERR, "dstTableQ size = ");
  ytFile_printComma_size_t(ytLib_ERR, mem >> 20);
  fprintf(ytLib_ERR, " [MB]\n");
  fflush(ytLib_ERR);

  /* The final result is stored in db->Q[Q_send]. */
  ytPCGraph * g = NULL;
  if(args->MPI_rank == 0) g = para9_graph(p, db->Q[Qb.send], db);

  return g;
}/* parallel9() ============================================================ */

/* End of file */
