/*
  ytDoubleArray.{h,c} : Integer Array
  Copyright (C) 2016, Yoshinori Tamada <tamada  A T ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "ytDoubleArray.h"
#include <assert.h>

/*!\brief Integer Array
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytDoubleArray */

#define YTLIB_INT_ARRAY_DEFAULT_EXPAND_SIZE 16

int ytDoubleArray_cmpAsc(const void * aa, const void * bb){
  const double * a = aa;
  const double * b = bb;

  if(*a < *b){
    return -1;
  } else if(*b < *a){
    return 1;
  } else {
    return 0;
  }
}/* ytDoubleArray_cmpAsc() ================================================= */

void ytDoubleArray_init(ytDoubleArray * this){
  this->ptr = NULL;
  this->size = 0;
  this->buff_size = 0;
}/* ytDoubleArray_init() =================================================== */

ytDoubleArray * ytDoubleArray_new(){
  ytDoubleArray * this = malloc(sizeof(ytDoubleArray));
  ytDoubleArray_init(this);
  return this;
}/* ytDoubleArray_new() ==================================================== */

void ytDoubleArray_delete(ytDoubleArray * this){
  free(this->ptr);
  free(this);
}/* ytDoubleArray_delete() ================================================= */

ytDoubleArray * ytDoubleArray_arrayNew(size_t size){
  ytDoubleArray * this = malloc(sizeof(ytDoubleArray) * size);
  for(size_t i = 0; i < size; i++){
    ytDoubleArray_init(this + i);
  }
  return this;
}/* ytDoubleArray_arrayNew() =============================================== */

size_t ytDoubleArray_size(const ytDoubleArray * this){
  return this->size;
}/* ytDoubleArray_size() =================================================== */

/*!\brief Changes the size of buffer.
 *
 *\param this
 *\param new_size
 *
 *\return \a this, or NULL if failed.
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytDoubleArray */
ytDoubleArray * ytDoubleArray_setBuffSize(ytDoubleArray * this, size_t new_size){
  double * new_ptr = realloc(this->ptr, sizeof(double) * new_size);

  if(new_ptr != NULL){
    if(new_size < this->size){
      this->size = new_size;
    }
    this->buff_size = new_size;
    this->ptr = new_ptr;
  } else {
    return NULL;
  }
  return this;
}/* ytDoubleArray_setBuffSize() ============================================ */

ytDoubleArray * ytDoubleArray_setSize(ytDoubleArray * this, size_t new_size){
  if(this->buff_size < new_size){
    ytDoubleArray * ret = ytDoubleArray_setBuffSize(this, new_size);
    if(ret == NULL) return NULL;
  }
  this->size = new_size;
  return this;
}/* ytDoubleArray_setSize() ================================================ */

void ytDoubleArray_clear(ytDoubleArray * this){
  this->size = 0;
}/* ytDoubleArray_clear() ================================================== */

void ytDoubleArray_add(ytDoubleArray * this, double value){
  if(this->buff_size == this->size){
    size_t new_size;
    if(this->buff_size == 0){
      new_size = YTLIB_INT_ARRAY_DEFAULT_EXPAND_SIZE;
    } else {
      new_size = this->buff_size * 2;
    }
#ifdef NDEBUG
    ytDoubleArray_setBuffSize(this, new_size);
#else
    assert(ytDoubleArray_setBuffSize(this, new_size) != NULL);
#endif
  }
  this->ptr[this->size++] = value;
}/* ytDoubleArray_add() ==================================================== */

void ytDoubleArray_set(ytDoubleArray * this, size_t index, double value){
  if(this->size <= index){
    ytDoubleArray_setSize(this, index + 1);
  }
  this->ptr[index] = value;
}

double ytDoubleArray_get(const ytDoubleArray * this, size_t index){
  assert(index < this->size);
  return this->ptr[index];
}/* ytDoubleArray_get() =================================================== */

void ytDoubleArray_sort(ytDoubleArray * this){
  qsort(this->ptr, this->size, sizeof(int), ytDoubleArray_cmpAsc);
}/* ytDoubleArray_sort() =================================================== */

void ytDoubleArray_print(ytDoubleArray * this, FILE * fp, char * delim){
  if(delim == NULL) delim = ", ";
  if(this->size != 0) fprintf(fp, " %f", this->ptr[0]);
  for(size_t i = 1; i < this->size; i++){
    fprintf(fp, "%s%f", delim, this->ptr[i]);
  }
}/* ytDoubleArray_print() ================================================== */

/* End of file. */
