/*
  ytIntArray.{h,c} : Integer Array
  Copyright (C) 2016, Yoshinori Tamada <tamada  A T ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "ytIntArray.h"
#include <assert.h>
#include <string.h>

/*!\brief Integer Array
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytIntArray */

#define YTLIB_INT_ARRAY_DEFAULT_EXPAND_SIZE 16

int ytIntArray_cmpAsc(const void * aa, const void * bb){
  const int * a = aa;
  const int * b = bb;

  if(*a < *b){
    return -1;
  } else if(*b < *a){
    return 1;
  } else {
    return 0;
  }
}/* ytIntArray_cmpAsc() ==================================================== */

void ytIntArray_init(ytIntArray * this){
  this->ptr = NULL;
  this->size = 0;
  this->buff_size = 0;
}/* ytIntArray_init() ====================================================== */

ytIntArray * ytIntArray_new(){
  ytIntArray * this = malloc(sizeof(ytIntArray));
  ytIntArray_init(this);
  return this;
}/* ytIntArray_new() ======================================================= */

void ytIntArray_delete(ytIntArray * this){
  free(this->ptr);
  free(this);
}/* ytIntArray_delete() ==================================================== */

ytIntArray * ytIntArray_arrayNew(size_t size){
  ytIntArray * this = malloc(sizeof(ytIntArray) * size);
  for(size_t i = 0; i < size; i++){
    ytIntArray_init(this + i);
  }
  return this;
}/* ytIntArray_arrayNew() ================================================== */

size_t ytIntArray_size(const ytIntArray * this){
  return this->size;
}/* ytIntArray_size() ==================================================== */

/*!\brief Returns the current buffer size allocated.
 *\param this  pointer to the ytIntArray instance.
 *\return the size of allocated buffer in bytes.
 *\memberof ytIntArray */
size_t ytIntArray_buffSize(const ytIntArray * this){
  return this->buff_size;
}/* ytIntArray_buffSize() ================================================== */

/*!\brief Changes the size of buffer.
 *
 *\param this
 *\param new_size
 *
 *\return \a this, or NULL if failed.
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytIntArray */
ytIntArray * ytIntArray_setBuffSize(ytIntArray * this, size_t new_size){
  int * new_ptr = realloc(this->ptr, sizeof(int) * new_size);

  if(new_ptr != NULL){
    if(new_size < this->size){
      this->size = new_size;
    }
    this->buff_size = new_size;
    this->ptr = new_ptr;
  } else {
    return NULL;
  }
  return this;
}/* ytIntArray_setBuffSize() =============================================== */

/*!\brief Returns the size of memory where the specified instance occupies.
 *\param this pointer to the ytIntArray instance.
 *\return the size of occupied memory size by the instance.
 *\memberof ytIntArray */
size_t ytIntArray_memSize(const ytIntArray * this){
  size_t mem = sizeof(ytIntArray);
  mem += ytIntArray_buffSize(this);
  return mem;
}/* ytIntArray_memSize() ================================================= */

/*!\brief Returns the total size of memory used by the array of the array.
 *\memberof ytIntArray */
size_t ytIntArray_arrayMemSize(const ytIntArray * this, size_t n){
  size_t mem = 0;
  for(size_t i = 0; i < n; i++){
    mem += ytIntArray_memSize(this + i);
  }
  return mem;
}/* ytIntArray_arrayMemSize() ============================================ */

void ytIntArray_clear(ytIntArray * this){
  this->size = 0;
}/* ytIntArray_clear() =================================================== */

void ytIntArray_add(ytIntArray * this, int value){
  if(this->buff_size == this->size){
    size_t new_size;
    if(this->buff_size == 0){
      new_size = YTLIB_INT_ARRAY_DEFAULT_EXPAND_SIZE;
    } else {
      new_size = this->buff_size * 2;
    }
#ifdef NDEBUG
    ytIntArray_setBuffSize(this, new_size);
#else
    assert(ytIntArray_setBuffSize(this, new_size) != NULL);
#endif
  }
  this->ptr[this->size++] = value;
}/* ytIntArray_add() ===================================================== */

void ytIntArray_addIntArray(ytIntArray * this, const ytIntArray * values){
  if(this->buff_size < this->size + values->size){
    size_t new_size = this->size + values->size;
#ifdef NDEBUG
    ytIntArray_setBuffSize(this, new_size);
#else
    assert(ytIntArray_setBuffSize(this, new_size) != NULL);
#endif
  }
  memcpy(this->ptr + this->size, values->ptr, sizeof(int) * values->size);
  this->size += values->size;
}/* ytIntArray_addIntArray() =========================================== */

void ytIntArray_set(ytIntArray * this, size_t index, int value){

  while(this->buff_size <= index){
    size_t new_size;
    if(this->buff_size == 0){
      new_size = YTLIB_INT_ARRAY_DEFAULT_EXPAND_SIZE;
    } else {
      new_size = this->buff_size * 2;
    }
#ifdef NDEBUG
    ytIntArray_setBuffSize(this, new_size);
#else
    assert(ytIntArray_setBuffSize(this, new_size) != NULL);
#endif
  }
  this->ptr[index] = value;
  if(this->size < index + 1){
    this->size = index + 1;
  }
}/* ytIntArray_set() ===================================================== */

void ytIntArray_setSize(ytIntArray * this, size_t size){
  if(this->size < size){
    ytIntArray_setBuffSize(this, size);
  }
  this->size = size;
}/* ytIntArray_setSize() ================================================= */

void ytIntArray_copy(ytIntArray * this, size_t index, const int * array, 
                       size_t size){
  size_t newSize = index + size;
  if(this->buff_size < newSize){
    ytIntArray_setBuffSize(this, newSize);
  }
  memcpy(this->ptr + index, array, sizeof(int) * size);
  if(this->size < newSize) this->size = newSize;
}/* ytIntArray_copy() ==================================================== */

int ytIntArray_get(const ytIntArray * this, size_t index){
  assert(index < this->size);
  return this->ptr[index];
}/* ytIntArray_get() ===================================================== */

void ytIntArray_sort(ytIntArray * this){
  qsort(this->ptr, this->size, sizeof(int), ytIntArray_cmpAsc);
}/* ytIntArray_sort() ==================================================== */

void ytIntArray_print(const ytIntArray * this, FILE * fp, char * delim){
  if(delim == NULL) delim = ", ";
  if(this->size != 0) fprintf(fp, "%d", (this->ptr[0]));
  for(size_t i = 1; i < this->size; i++){
    fprintf(fp, "%s""%d", delim, (this->ptr[i]));
  }
}/* ytIntArray_print() =================================================== */

/* End of file. */
