/*
  util/ytParseArgs.{h,c} : Parsing command line arguments
  Copyright (C) 2016, Yoshinori Tamada <tamada  A T  ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include "../ytLib.h"
#include "ytParseArgs.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#define ytParseArgs_BUFF_SIZE 256

/*!\brief Checks if the character is a number character.
 *
 *\param ch  Character to be checked.
 *\param return Non-zero if \a ch is a number character.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\internal
 *\memberof ytParseArgs */
int ytParseArgs_isNumCh(char ch){
  int ret = 0;
  switch(ch){
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9':
	ret = 1;
	break;
  }
  return ret;
}
/* ytParseArgs_isNumCh() =================================================== */

/*!\brief Checks if the string is a number.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\internal
 *\memberof ytParseArgs */
int ytParseArgs_isNum(const char * str){
  while(*str != '\0'){
	switch(*str){
	case '0': case '1': case '2': case '3': case '4': 
	case '5': case '6': case '7': case '8': case '9':
	case '.': case '-':
	  str++;
	  break;

	default:
	  return 0;
	}
  }
  return 1;
}
/* ytParseArgs_isNum() ===================================================== */

int ytParseArgs_checkIntRegion(const char * str){
  while(*str != '\0'){
	switch(*str){
	case '0': case '1': case '2': case '3': case '4': 
	case '5': case '6': case '7': case '8': case '9':
	case '-': case ':':
	  str++;
	  break;
	  
	default:
	  return 0;
	}
  }
  return 1;
}
/* ytParseArgs_checkIntRegion() ============================================ */

/*
 *\return 0 if succeeded, or non-zero otherwise.
 */
int ytParseArgs_parseIntRegionPart(int ** ptr, int * len, char * value){
  int buff[ytParseArgs_BUFF_SIZE];
  char * p1 = strchr(value, '-');
  char * p2 = strchr(value, ':');
  
  int inc = 1;
  int ed = 1;
  int st = 1;
  int c = 0;  /* # of added elements in 'buff' */

  if(ytParseArgs_checkIntRegion(value) == 0) return 1;

  /* Starting value */
  st = atoi(value); 

  /* Ending value */
  if(p1 != NULL){
	ed = atoi(p1 + 1);
	if(ed < st) ed = st;
  } else {
	ed = st;
  }

  /* Increment value */
  if(p2 != NULL) inc = atoi(p2 + 1);
  
  while(st <= ed && c < ytParseArgs_BUFF_SIZE){
	buff[c++] = st;
	st += inc;
  }
  *ptr = realloc(*ptr, sizeof(int) * (*len + c));
  memcpy(*ptr + *len, buff, sizeof(int) * c);
  *len += c;

  return 0;
}
/* ytParseArgs_parseIntRegionPart() ======================================== */

/*!\brief Parses x[-y[:z]],... style format integer region.
 *
 *\param ptr  Pointer to the integer array to be set.
 *\param len  Length of the array.
 *\param value  x[-y[:z]],... style format string.
 *
 *\return 0 if succeeded, or non-zero if error found.
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *!\internal
 *\memberof ParseArgs */
int ytParseArgs_parseIntRegion(int ** ptr, int * len, char * value){

  while(1) {
	char * comma_pos = strchr(value, ',');
	if(comma_pos != NULL) *comma_pos = '\0';
	int r = ytParseArgs_parseIntRegionPart(ptr, len, value);
	if(r != 0) return 1;
	value = comma_pos;
	if(value == NULL) break;
	value++;
  }
  return 0;
}
/* ytParseAgs_parseIntRegion() ============================================= */

void ytParseArgs_parseStringArray(char *** ar, int * len, char * value){
  char ** new_ar = realloc(*ar, sizeof(char **) * (*len + 1));
  new_ar[*len] = value;
  (*len)++;
  *ar = new_ar;
}
/* ParseArgs_parse_string_array() ========================================== */

/*!\brief Parses a pair of the key and its value and set to the variable.
 *
 *\param argset  List of the acceptable arguments.  The field ParseArgs::key
 *    of the last element needs to be NULL to represent the end of the array.
 *\param key     argument key.
 *\param value   argument value.  
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\internal
 *\memberof ytParseArgs */
int ytParseArgs_parse(ytParseArgs * argset, char * key, char * value){
  /* Search 'key' from argset */
  int i = 0;
  int pos = -1;
  while(argset[i].key != NULL){
	if(strcmp(argset[i].key, key) == 0){
	  pos = i;  break;
	}
	i++;
  }
  if(pos == -1)	return -1; /* Key not found */

  switch(argset[i].type){
  case ytParseArgs_INT:
	{
	  if(value == NULL) return -2; /* Value not found */
	  char * endptr;
	  int v = strtol(value, &endptr, 10);
	  if(*endptr != '\0') return -5; /* parse error */
	  *((int *) argset[i].arg1) = v;
	}
	break;

  case ytParseArgs_UINT:
	{
	  if(value == NULL) return -2; /* Value not found */
	  char * endptr;
	  unsigned int v = strtoul(value, &endptr, 10);
	  if(*endptr != '\0') return -5; /* parse error */
	  *((unsigned int *) argset[i].arg1) = v;
	}
	break;

  case ytParseArgs_DOUBLE:
	{
	  if(value == NULL) return -2; /* Value not found */
	  char * endptr;
	  double v = strtof(value, &endptr);
	  if(*endptr != '\0') return -5; /* double parse error */
	  *((double *) argset[i].arg1) = v;
	}
	break;

  case ytParseArgs_STRING:
	if(value == NULL) return -2; /* Value not found */
	*((char **) argset[i].arg1) = value;
	break;

  case ytParseArgs_STRING_ARRAY:
	if(value == NULL) return -2; /* Value not found */
	ytParseArgs_parseStringArray((char ***) argset[i].arg1, 
								 (int *) argset[i].arg2, value);
	break;

  case ytParseArgs_ON_OFF:
	if(value == NULL) return -2; /* Value not found */
	if(strcasecmp(value, "on") == 0 ||
	   strcasecmp(value, "1") == 0 ||
	   strcasecmp(value, "true") == 0 ||
	   strcasecmp(value, "t") == 0 ||
	   strcasecmp(value, "yes") == 0){
	  *((int *) argset[i].arg1) = 1;
	} else if(strcasecmp(value, "off") == 0 ||
			  strcasecmp(value, "0") == 0 ||
			  strcasecmp(value, "false") == 0 ||
			  strcasecmp(value, "f") == 0 ||
			  strcasecmp(value, "no") == 0){
	  *((int *) argset[i].arg1) = 0;
	} else {
	  return -4; /* On|Off parse error */
	}
	break;

  case ytParseArgs_ON:
	*((int *) argset[i].arg1) = 1;
	if(value != NULL) return -3; /* value not necessary */
	break;

  case ytParseArgs_INT_REGION:
	{
	  if(value == NULL) return -2; /* Value not found */
	  int r = ytParseArgs_parseIntRegion((int **) argset[i].arg1,
										 (int *) argset[i].arg2, value);
	  if(r != 0) return -6; /* parse error */
	}
	break;

  case ytParseArgs_FUNC:
	((void (*) (void)) argset[i].arg1)();
	break;
  }/* switch() */
  return 0;
}
/* ytParseArgs_parse() ===================================================== */

void ytParseArgs_error(int r, char * key){
  switch(r){
  case -1: break; /* ignore */
  case -2:
	fprintf(ytLib_ERR, "ERROR: \"%s\" requires its argument.\n", key);
	break;
  default:
	fprintf(ytLib_ERR, "ERROR: Argument parsing failed.\n");
	break;
  }
}
/* ytParseArgs_error() ======================================================= */

/*\brief Checks if the given string is a key.
 *
 * A string that begins with '-' (minus), followed by
 * non-number letters are recognized as a key of an argument value.
 *
 *\return non zero value if \a str is a key, or 0 otherwise.
 *\internal
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytParseArgs */
int ytParseArgs_isKey(char * str){
  if(str == NULL) return 0;
  if(str[0] == '-' && strcmp(str, "-") != 0 && !ytParseArgs_isNumCh(str[1])){
	return 1;
  } else {
	return 0;
  }
}
/* ytParseArgs_isKey() ===================================================== */

/*!\brief Parsing command line arguments.
 *
 *\param argset Array of ParseArgs structure.  The ParseArgs::key field of the
 *   terminating element needs to be NULL to represent the end of array.
 *\param argc   Number of arguments in \a argv, which is passed by main(). 
 *\param argv   Array of arguments passed by main().
 *
 *\return 0 if succeeded, otherwise the index of \a argv that caused the error.
 *
 *
 *<DIV align="right"><A href="#top">Go to Top</A></DIV>
 *\memberof ytParseArgs */
int ytParseArgs_set(ytParseArgs * argset, int argc, char ** argv){
  
  int i;
  int flag = 0;
  char * key = NULL;
  
  for(i = 1; i < argc; i++){

	if(flag){
	  /* argv[i] is a value without a key */
	  int r = ytParseArgs_parse(argset, ytParseArgs_DEFAULT_KEY, argv[i]);
	  if(r != 0) ytParseArgs_error(r, "");
	  return i;
	} else if(strcmp(argv[i], "-") == 0){
	  /* All remaining arguments are values of the default key. */
	  flag = 1;

	} else if(ytParseArgs_isKey(argv[i])){
	  /* argv[i] is a key */
	  if(key != NULL){
		int r = ytParseArgs_parse(argset, key, NULL);
		if(r != 0){
		  ytParseArgs_error(r, key);
		  return i;
		}
	  }
	  key = argv[i];
	  
	} else if(key != NULL){
	  /* argv[i] is a value with a key. */
	  int r = ytParseArgs_parse(argset, key, argv[i]);
	  if(r == -3){
		/* Value is not used */
		if(ytParseArgs_isKey(argv[i])){
		  key = argv[i];
		} else {
		  key = NULL;
		  i--;
		}
	  } else if(r == -1){
		return i - 1;
	  } else if(r != 0){
		ytParseArgs_error(r, key);
		return i;
	  } else {
		key = NULL;
	  }
	} else {
	  /* argv[i] is a value without a key. */
	  int r = ytParseArgs_parse(argset, ytParseArgs_DEFAULT_KEY, argv[i]);
	  if(r != 0){
		ytParseArgs_error(r, ytParseArgs_DEFAULT_KEY);
		return i;
	  }
	}
  }/* for i < args */

  if(key != NULL){
	int r = ytParseArgs_parse(argset, key, NULL);
	if(r != 0){
	  ytParseArgs_error(r, key);
	  return i - 1;
	}
  }
  
  return 0; /* no error */
}
/* ytParseArgs_set() ======================================================= */

/* End of file.  */
