/*
  util/ytParseArgs.{h,c} : Parsing command line arguments
  Copyright (C) 2016, Yoshinori Tamada <tamada  A T  ytlab.jp>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright 
    notice, this list of conditions and the following disclaimer in 
    the documentation and/or other materials provided with the 
    distribution.

  * Neither the name of The University of Tokyo nor the names of its 
    contributors may be used to endorse or promote products derived 
    from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __YTLIB_PARSE_ARGS_H
#define __YTLIB_PARSE_ARGS_H

/*\defgroup ytParseArgs */

/*!\enum ytParseArgsType
 *\memberof ytParseArgs
 *\brief Supported value types.
 *
 * Specify one of these as the second field of ParseArgs structure.
 *
 *\file ytParseArgs.h
 */
typedef enum {
  /*!\brief Integer value (<tt>int</tt>).  */
  ytParseArgs_INT,

  /*!\brief Unsigned integer values (<tt>unsigned int</tt>). */
  ytParseArgs_UINT,
  
  /*!\brief Double precision floating value (<tt>double</tt>). */
  ytParseArgs_DOUBLE,
  
  /*!\brief Characer string value (<tt>char *</tt>). */
  ytParseArgs_STRING,

  /*!\brief Character string arrays */
  ytParseArgs_STRING_ARRAY,

  /*!\brief Integer value 1 is set if the key is found.  
    This does not take its value. */
  ytParseArgs_ON, 

  /*!\brief Integer value 1 is set if its value is one of <tt>on</tt>, <tt>1</tt>, and <tt>true</tt>. */
  ytParseArgs_ON_OFF,

  /*!\brief Array of integer values (<tt>int *</tt>).
   * The number of elements is set to the optional second integer pointer (<tt>int</tt>). */
  ytParseArgs_INT_REGION,

  /*!\brief Executes a function <tt>(void) f(void)</tt>. */
  ytParseArgs_FUNC,
  
} ytParseArgsType;

/*!\brief Structure for parsing command line arguments.
 *
 * Specify the array of this structure as the first argument
 * of ytParseArgs_set().  The field ParseArgs::key of the last element of the array
 * has to be NULL to represent the end of the array.
 *
 *\section example Example
 *
 *\code
 * int main(int argc, char * argv[]){
 *   int v = 0;
 *   char * s = "not set";
 *   ytParseArgs argset[] = {
 *     {"-v", ytParseArgs_INT,    &v},
 *     {"-s", ytParseArgs_STRING, &c},
 *     {NULL},                       // The last element has to be {NULL}.
 *   };
 *
 *   int r = ytParseArgs_set(argset, argc, argv);
 *   if(r > 0){
 *      printf("Parsing Error: %s\n", argv[r]);
 *   }
 *   // If "-v 10" is specified, 10 is set to "v".
 *   // Otherwise, v == 0.
 *   printf("v = %d\n", v);
 *
 *   // If "-s foobar" is specified, "foobar" is set to "s".
 *   // Otherwise, s == "not set".
 *   printf("s = %s\n", s);
 *   return 0;
 * }\endcode
 *
 *\author Yoshinori Tamada &lt;tamada at ims.u-tokyo.ac.jp&gt;
 */
typedef struct {
  /*!\brief Argument key that begins with &quot;-&quot; (hyphen). */
  char * key;

  /*!\brief Argument value type. */
  ytParseArgsType type;

  /*!\brief Pointer to which the parsed value is set. */
  void * arg1;

  /*!\brief Optional pointer used depending on the type of a value. */
  void * arg2;
} ytParseArgs;

#ifdef DOXY
/*!\brief Default key for an argument without a key.
 *
 *\memberof ParseArgs */
#define ytParseArgs_DEFAULT_KEY
#else
#define ytParseArgs_DEFAULT_KEY  "__ARGS_WITHOUT_KEY__"
#endif

int ytParseArgs_set(ytParseArgs * argset, int argc, char ** argv);

#endif /* __YTLIB_PARSE_ARGS_H */

/* End of file. */
